# -*- coding: utf-8 -*-
"""
Created on Mon Jul 17 13:08:34 2023

@author: veenstra
"""

import os
import pytest
import dfm_tools as dfmt
import numpy as np
import pandas as pd
import datetime as dt
import xarray as xr
import shapely
import geopandas as gpd
from dfm_tools.interpolate_grid2bnd import (tidemodel_componentlist,
                                            components_translate_upper,
                                            interp_regularnc_to_plipointsDataset,
                                            check_time_extent,
                                            ext_add_boundary_object_per_polyline,
                                            ds_apply_conventions,
                                            ds_apply_conversion_dict,
                                            open_prepare_dataset,
                                            _ds_sel_time_outside,
                                            )
from dfm_tools.hydrolib_helpers import get_ncbnd_construct
import hydrolib.core.dflowfm as hcdfm
from dfm_tools.errors import OutOfRangeError
import warnings


@pytest.mark.unittest
def test_check_time_extent(cmems_dataset_4times):
    ds = cmems_dataset_4times
    
    # prior to ds timerange
    try:
        check_time_extent(ds, tstart="2011-01-01", tstop="2011-01-01")
    except OutOfRangeError as e:
        assert "tstart" in str(e)
    
    # after ds timerange
    try:
        check_time_extent(ds, tstart="2021-01-01", tstop="2021-01-01")
    except OutOfRangeError as e:
        assert "tstop" in str(e)
    
    # in ds timerange
    check_time_extent(ds, tstart="2020-01-01", tstop="2020-01-01")
    
    # on ds timerange
    check_time_extent(ds, tstart="2019-12-31 12:00:00", tstop="2020-01-03 12:00:00")


@pytest.mark.unittest
def test_conversion_dict():
    """
    since notations of equations are sometimes updated, convenient to have this test
    that checks the conversion values that were once used.
    """
    dict_keys_waq = ['tracerbndOXY', 'tracerbndNO3', 'tracerbndPO4', 'tracerbndSi', 
                     'tracerbndPON1', 'tracerbndPOP1', 'tracerbndPOC1', 
                     'tracerbndDON', 'tracerbndDOP', 'tracerbndDOC', 'tracerbndOpal']
    
    conversion_expected = {'tracerbndOXY': 0.032,
     'tracerbndNO3': 0.014,
     'tracerbndPO4': 0.03097,
     'tracerbndSi': 0.02808,
     'tracerbndPON1': 0.004226415094339623,
     'tracerbndPOP1': 0.0005843396226415094,
     'tracerbndPOC1': 0.024,
     'tracerbndDON': 0.013693584905660377,
     'tracerbndDOP': 0.0005843396226415094,
     'tracerbndDOC': 0.11678671698113208,
     'tracerbndOpal': 0.0018252}
    
    conversion_dict = dfmt.get_conversion_dict()
    for quan in dict_keys_waq:
        assert np.abs(conversion_dict[quan]['conversion'] - conversion_expected[quan]) < 1e-9


@pytest.mark.unittest
def test_tidemodel_componentlist():
    comp_list = tidemodel_componentlist(tidemodel='FES2014', convention=False)
    comp_list_convention = tidemodel_componentlist(tidemodel='FES2014', convention=True)
    
    assert len(comp_list) == 34
    assert len(comp_list_convention) == 34
    assert comp_list[1] == 'eps2'
    assert comp_list_convention[1] == 'EPSILON2'


@pytest.mark.unittest
def test_components_translate_upper():
    comp_list = components_translate_upper(['m2','eps2','e2'])
    assert comp_list == ['M2','EPSILON2','EPSILON2']


@pytest.mark.unittest
@pytest.mark.requireslocaldata
def test_plipointsDataset_fews_accepted():
    """
    check if FEWS netcdf bnd export is correctly processed to a 
    hcdfm ForcingModel object including the necessary conversions
    """
    file_nc_fews = r'p:\dflowfm\maintenance\JIRA\06000-06999\06187\C01\salinity_DCSM-FM_OB_all.nc'
    data_interp = xr.open_dataset(file_nc_fews)
    
    #convert plipointsDataset to hydrolib ForcingModel
    forcingmodel_object = dfmt.plipointsDataset_to_ForcingModel(plipointsDataset=data_interp)
    
    forcing0 = forcingmodel_object.forcing[0]
    assert isinstance(forcingmodel_object, hcdfm.ForcingModel)
    assert isinstance(forcing0, hcdfm.T3D)
    assert forcing0.quantityunitpair[1].unit == 'ppt'
    
    # test whether so was renamed to salinitybnd
    assert forcing0.quantityunitpair[1].quantity == 'salinitybnd'


@pytest.mark.systemtest
def test_plipointsDataset_to_ForcingModel_drop_allnan_points():
    #construct polyfile gdf
    point_x = [-71.5, -71.5, -71.5, -71.5,]
    point_y = [12.5, 12.6, 12.7, 12.8]
    polyobject_pd = pd.DataFrame(dict(x=point_x, y=point_y))
    poly_object = dfmt.DataFrame_to_PolyObject(polyobject_pd, name="abc_bnd")
    polyfile_object = hcdfm.PolyFile()
    polyfile_object.objects.append(poly_object)
    gdf_points = dfmt.PolyFile_to_geodataframe_points(polyfile_object)
    
    # actual cmems data
    no3_values = [[[       np.nan,        np.nan,        np.nan,        np.nan,
             5.2622618e-05],
            [       np.nan,        np.nan,        np.nan,        np.nan,
             5.1448391e-05],
            [9.0356334e-05, 1.8063841e-04,        np.nan, 7.0045113e-05,
             4.9546972e-05],
            [4.0657567e-05, 4.6050434e-05, 4.7762936e-05, 4.2734890e-05,
             4.0186114e-05],
            [2.6492798e-05, 2.8967228e-05, 3.0298394e-05, 3.1432221e-05,
             3.2138258e-05],
            [2.4448847e-05, 2.6705173e-05, 2.6929094e-05, 2.5548252e-05,
             2.5827681e-05]]]
    
    ds = xr.Dataset()
    ds['longitude'] = xr.DataArray([-72.  , -71.75, -71.5 , -71.25, -71.  ], dims='longitude')
    ds['latitude'] = xr.DataArray([12.  , 12.25, 12.5 , 12.75, 13.  , 13.25], dims='latitude')
    ds['time'] = xr.DataArray([639204.],dims='time').assign_attrs({'units': 'hours since 1950-01-01'})
    ds['tracerbndNO3'] =  xr.DataArray(no3_values, dims=('time','latitude','longitude')).assign_attrs({'units': 'mmol m-3'})
    ds = xr.decode_cf(ds)
    
    # interpolate to points and convert to forcingmodel
    data_interp = interp_regularnc_to_plipointsDataset(data_xr_reg=ds, gdf_points=gdf_points)
    forcingmodel_object = dfmt.plipointsDataset_to_ForcingModel(plipointsDataset=data_interp)
    
    # check if the resulting forcingmodel does not have 4 but 2 points
    # the first two points were skipped because they were all nan
    assert len(forcingmodel_object.forcing) == 2
    assert forcingmodel_object.forcing[0].name == 'abc_bnd_0003'
    assert forcingmodel_object.forcing[1].name == 'abc_bnd_0004'


def test_ds_sel_time_outside(cmems_dataset_4times):
    ds = cmems_dataset_4times
    
    # exact outer bounds
    tstart = "2019-12-31 12:00"
    tstop = "2020-01-03 12:00"
    ds_sel = _ds_sel_time_outside(ds, tstart, tstop)
    assert ds_sel.time.values[0] <= pd.Timestamp(tstart)
    assert ds_sel.time.values[1] > pd.Timestamp(tstart)
    assert ds_sel.time.values[-2] < pd.Timestamp(tstop)
    assert ds_sel.time.values[-1] >= pd.Timestamp(tstop)
    assert len(ds_sel.time) == 4
    
    # exact inner bounds
    tstart = "2020-01-01 12:00"
    tstop = "2020-01-02 12:00"
    ds_sel = _ds_sel_time_outside(ds, tstart, tstop)
    assert ds_sel.time.values[0] <= pd.Timestamp(tstart)
    assert ds_sel.time.values[1] > pd.Timestamp(tstart)
    assert ds_sel.time.values[-2] < pd.Timestamp(tstop)
    assert ds_sel.time.values[-1] >= pd.Timestamp(tstop)
    assert len(ds_sel.time) == 2

    # inexact inner bounds
    tstart = "2020-01-01"
    tstop = "2020-01-03"
    ds_sel = _ds_sel_time_outside(ds, tstart, tstop)
    assert ds_sel.time.values[0] <= pd.Timestamp(tstart)
    assert ds_sel.time.values[1] > pd.Timestamp(tstart)
    assert ds_sel.time.values[-2] < pd.Timestamp(tstop)
    assert ds_sel.time.values[-1] >= pd.Timestamp(tstop)
    assert len(ds_sel.time) == 4
    
    # tstart out of bounds
    tstart = "2019-12-30 12:00"
    tstop = "2020-01-03 12:00"
    with pytest.raises(OutOfRangeError) as e:
        _ds_sel_time_outside(ds, tstart, tstop)
    assert "requested tstart 2019-12-30 12:00:00 outside" in str(e.value)
    
    # tstart out of bounds
    tstart = "2019-12-31 12:00"
    tstop = "2030-01-03 12:00"
    with pytest.raises(OutOfRangeError) as e:
        _ds_sel_time_outside(ds, tstart, tstop)
    assert "requested tstop 2030-01-03 12:00:00 outside" in str(e.value)


@pytest.mark.systemtest
def test_open_prepare_dataset_correctdepths(tmp_path, cmems_dataset_4times):
    """
    to validate open_prepare_dataset behaviour for depths, in the past the depth values got lost and replaced by depth idx
    """
    
    ds_moretime = cmems_dataset_4times
    file_nc = tmp_path / 'temp_cmems_dummydata.nc'
    ds_moretime.to_netcdf(file_nc)
    
    ds_moretime_import = open_prepare_dataset(dir_pattern=file_nc, quantity='salinitybnd', tstart='2020-01-01 12:00', tstop='2020-01-02 12:00')
    assert len(ds_moretime_import.time) == 2


@pytest.mark.unittest
def test_ds_apply_conventions(cmems_dataset_4times):
    # generate datset with depths defined positive down
    ds_moretime = cmems_dataset_4times
    ds_moretime['depth'] = -1 * ds_moretime['depth']
    ds_moretime['depth'].attrs['positive'] = 'down'
    ds_converted = ds_apply_conventions(data_xr=ds_moretime)
    
    ncbnd_construct = get_ncbnd_construct()
    varn_depth = ncbnd_construct['varn_depth']
    dimn_depth = ncbnd_construct['dimn_depth']
    assert varn_depth in ds_converted.variables
    assert dimn_depth in ds_converted.dims
    
    depth_expected = np.array([-0.494025, -1.541375, -2.645669, -3.819495, -5.078224])
    depth_actual = ds_converted[varn_depth].to_numpy()
    assert np.allclose(depth_actual, depth_expected)
    

@pytest.mark.unittest
def test_ds_apply_conversion_dict_rename(cmems_dataset_4times):
    conversion_dict = dfmt.get_conversion_dict()
    ds_moretime = cmems_dataset_4times
    ds_converted = ds_apply_conversion_dict(data_xr=ds_moretime, conversion_dict=conversion_dict, quantity='salinitybnd')
    assert 'so' in ds_moretime.data_vars
    assert 'salinitybnd' in ds_converted.data_vars
    assert np.allclose(ds_converted['salinitybnd'], ds_moretime['so'], equal_nan=True)


@pytest.mark.unittest
def test_ds_apply_conversion_dict_rename_and_factor(cmems_dataset_4times):
    conversion_dict = dfmt.get_conversion_dict()
    ds_moretime = cmems_dataset_4times
    ds_moretime = ds_moretime.rename_vars({'so':'o2'})
    ds_moretime['o2'] = ds_moretime['o2'].assign_attrs({'units':'dummy'})
    ds_converted = ds_apply_conversion_dict(data_xr=ds_moretime, conversion_dict=conversion_dict, quantity='tracerbndOXY')
    assert 'o2' in ds_moretime.data_vars
    assert 'tracerbndOXY' in ds_converted.data_vars
    assert np.allclose(ds_converted['tracerbndOXY'], ds_moretime['o2']*0.032, equal_nan=True)
    assert ds_converted['tracerbndOXY'].attrs['units'] == 'g/m3'


@pytest.mark.unittest
def test_open_prepare_dataset_slightly_different_latlons(tmp_path, cmems_dataset_4times):
    """
    to check whether an error is raised when trying to combine datasets with slightly 
    different coordinates: https://github.com/Deltares/dfm_tools/issues/574
    
    """
    ds1 = cmems_dataset_4times.isel(time=slice(None,2))
    ds2 = cmems_dataset_4times.isel(time=slice(2,None))
    
    # deliberately alter longitude coordinate slightly
    ds_lon = ds1.longitude.to_numpy().copy()
    ds_lon[1] += 1e-8
    ds2['longitude'] = xr.DataArray(ds_lon,dims='longitude')
    
    file_nc1 = tmp_path / 'temp_cmems_2day_p1.nc'
    file_nc2 = tmp_path / 'temp_cmems_2day_p2.nc'
    ds1.to_netcdf(file_nc1)
    ds2.to_netcdf(file_nc2)
    file_nc = tmp_path / 'temp_cmems_2day_*.nc'
    
    with pytest.raises(ValueError) as e:
        open_prepare_dataset(file_nc, quantity='salinitybnd', tstart='2020-01-01 12:00:00', tstop='2020-01-02 12:00:00')
    
    # ValueError: cannot align objects with join='exact' where index/labels/sizes are not equal along these coordinates (dimensions): 'longitude' ('longitude',)
    assert "cannot align objects with join='exact' " in str(e.value)


@pytest.mark.unittest
def test_interp_regularnc_to_plipointsDataset(cmems_dataset_notime):
    """
    Linear interpolation to a new dimension in dfmt.interp_regularnc_to_plipoints() 
    resulted in unexpected nan values since scipy 1.10.0.
    More info in https://github.com/pydata/xarray/issues/7701 and related issues.
    Since this interpn is way faster than interp1d on a grid or all separate xy points, 
    we sticked to interpn but filled nan values with a nearest interpolation. 
    This only happens if there are non-nan values in the surrounding cells.
    In this test we check whether this combination of interpolation results in the same values as interp1d.
    This is mainly relevant when interpolating to existing lat/lon values, so that is the case we test here.
    This method gives as much valid values as possible given the input dataset, but does not fill nans where it should not do that.
    """
    
    ncbnd_construct = get_ncbnd_construct()
    dimn_point = ncbnd_construct['dimn_point']
    dimn_depth = ncbnd_construct['dimn_depth']
    varn_depth = ncbnd_construct['varn_depth']
    varn_pointname = ncbnd_construct['varn_pointname']
    
    ds = cmems_dataset_notime.copy()
    ds = ds.rename_dims({'depth':dimn_depth})
    ds = ds.rename_vars({'depth':varn_depth})
    so_np = ds['so'].to_numpy()
    lons = ds['longitude'].to_numpy()
    lats = ds['latitude'].to_numpy()
    
    for ipoint in range(3):
        x_xr = xr.DataArray([lons[ipoint]],dims=(dimn_point))
        y_xr = xr.DataArray([lats[ipoint]],dims=(dimn_point))
        
        # interp1d # these are actually irrelevant now, are not used for testing
        interp_with_floats = ds.interp(longitude=x_xr[0], latitude=y_xr[0], method='linear').so #selecting one value from the da drops the new plipoints dimension
        interp_with_da_existing = ds.interp(longitude=x_xr.values, latitude=y_xr.values, method='linear').so.isel(longitude=0,latitude=0) #using the DataArray values keeps lat/lon dimenions, gives the same interp result
        
        geom = shapely.points(x_xr, y_xr)
        gdf = gpd.GeoDataFrame(data={varn_pointname:[f'name_{ipoint+1:04d}']}, geometry=geom)
        interp_with_da_newdim = dfmt.interp_regularnc_to_plipointsDataset(ds, gdf, load=True)
        interp_da_actual = interp_with_da_newdim.so.isel({dimn_point:0})
        
        #define expected values since in some cases like point 0 this is not the same as interp1d returns
        interp_da_expected = xr.DataArray(so_np[:,ipoint,ipoint],dims=(dimn_depth))
        
        print(ipoint)
        print(interp_with_floats.to_numpy())
        print(interp_with_da_existing.to_numpy())
        print(interp_da_actual.to_numpy())
        print(interp_da_expected.to_numpy())
        
        assert (interp_with_floats.isnull()==interp_with_da_existing.isnull()).all()
        assert (interp_da_expected.isnull()==interp_da_actual.isnull()).all()
    
    """
    prints with scipy 1.11.3:
    0
    [35.819576 35.802788 35.786003       nan       nan]
    [35.819576 35.802788 35.786003       nan       nan]
    [35.819576 35.802788 35.786003 35.776848 35.781425]
    [35.819576 35.802788 35.786003 35.776848 35.781425]
    1
    [35.824154 35.810417 35.796684       nan       nan]
    [35.824154 35.810417 35.796684       nan       nan]
    [35.824154 35.810417 35.796684       nan       nan]
    [35.824154 35.810417 35.796684       nan       nan]
    2
    [35.82873 35.81805      nan      nan      nan]
    [35.82873 35.81805      nan      nan      nan]
    [35.82873 35.81805      nan      nan      nan]
    [35.82873 35.81805      nan      nan      nan]
    """


@pytest.mark.unittest
def test_interp_regularnc_to_plipointsDataset_checkvardimnames(cmems_dataset_notime):
    ncbnd_construct = get_ncbnd_construct()
    dimn_point = ncbnd_construct['dimn_point']
    dimn_depth = ncbnd_construct['dimn_depth']
    varn_depth = ncbnd_construct['varn_depth']
    varn_pointx = ncbnd_construct['varn_pointx']
    varn_pointy = ncbnd_construct['varn_pointy']
    varn_pointname = ncbnd_construct['varn_pointname']
    
    ds = cmems_dataset_notime.copy()
    ds = ds.rename_dims({'depth':dimn_depth})
    ds = ds.rename_vars({'depth':varn_depth})
    lons = ds['longitude'].to_numpy()
    lats = ds['latitude'].to_numpy()
    
    x_xr = xr.DataArray([lons[0]],dims=(dimn_point))
    y_xr = xr.DataArray([lats[0]],dims=(dimn_point))
    
    geom = shapely.points(x_xr, y_xr)
    gdf = gpd.GeoDataFrame(data={varn_pointname:['name_0001']}, geometry=geom)
    interp_with_da_newdim = dfmt.interp_regularnc_to_plipointsDataset(ds, gdf, load=True)
    
    # check if only expected dims/vars are present
    varn_inda = set(list(interp_with_da_newdim.variables))
    varn_expected = set(['so', varn_depth, varn_pointx, varn_pointy, varn_pointname])
    dimn_inda = set(list(interp_with_da_newdim.dims))
    dimn_expected = set([dimn_point, dimn_depth])
    assert varn_inda == varn_expected
    assert dimn_inda == dimn_expected


@pytest.mark.systemtest
@pytest.mark.requireslocaldata
def test_interpolate_tide_to_plipoints(data_dcsm_gdf):
    nanvalue = -999
    gdf_points = data_dcsm_gdf
    
    tidemodel_list = ['tpxo80_opendap', 'FES2014', 'FES2012', 'EOT20', 'GTSMv4.1']#, 'GTSMv4.1_opendap']
    for tidemodel in tidemodel_list:
        print(tidemodel)
        dtstart = dt.datetime.now()
        component_list_tidemodel = tidemodel_componentlist(tidemodel, convention=True)
        
        if tidemodel=='tpxo80_opendap': # 4.7 sec (all components: 5.8 sec)
            amp_expected = np.array([1.09643936, 1.08739412, 1.08555067])
            phs_expected = np.array([81.92059326171875, 82.513671875, 82.69258117675781])
        elif tidemodel=='FES2014': # 10.5 sec (all components: 97.6 sec)
            amp_expected = np.array([1.1147955656051636, 1.1004363298416138, 1.0878639221191406])
            phs_expected = np.array([81.8884048461914, 82.19799041748047, 82.39784240722656])
        elif tidemodel=='FES2012': # 9.5 sec (all components: 92.4 sec)
            amp_expected = np.array([nanvalue, 1.0839321613311768, 1.0718189477920532])
            phs_expected = np.array([nanvalue, 82.33390808105469, 82.60922241210938])
        elif tidemodel=='EOT20': # 9.9 sec (all components: 61.9 sec)
            amp_expected = np.array([nanvalue, 1.10231938,  1.08968516])
            phs_expected = np.array([nanvalue, 82.54157149, 82.76990983])
        elif tidemodel in ['GTSMv4.1','GTSMv4.1_opendap']: # 13.8 sec vs 39.2 sec (all components: 82.4 sec vs 498.4 sec)
            amp_expected = np.array([1.13028932, 1.10648024, 1.09396541])
            phs_expected = np.array([81.21875763, 81.41669464, 81.66479492])
        
        for component_list in [['M2','S2']]: # [None]: # 
            data_interp = dfmt.interpolate_tide_to_plipoints(tidemodel=tidemodel, gdf_points=gdf_points, component_list=component_list)
            
            compnames_now = data_interp['compno'].to_numpy().tolist()
            if component_list is None:
                compnames_expected = component_list_tidemodel
            else:
                compnames_expected = component_list
            
            data_interp_M2 = data_interp.sel(compno='M2').fillna(nanvalue)
            
            amp_now = data_interp_M2['amplitude'].to_numpy()
            phs_now = data_interp_M2['phase'].to_numpy()
            
            assert compnames_now == compnames_expected
            assert (np.abs(amp_expected-amp_now)<1e-6).all()
            assert (np.abs(phs_expected-phs_now)<1e-6).all()
        
        print(f'>> tide interpolation from {tidemodel} took: ',end='')
        print(f'{(dt.datetime.now()-dtstart).total_seconds():.2f} sec')

    
@pytest.mark.unittest
def test_interpolate_tide_to_forcingmodel(data_dcsm_gdf):
    """
    This tests adds to test_interpolate_tide_to_plipoints, since it also interpolates to ForcingModel
    Furthermore, it runs on Github since it does not depend on local data
    """
    
    tidemodel = 'tpxo80_opendap'
    component_list = ['M2']
    gdf_points = data_dcsm_gdf

    data_interp = dfmt.interpolate_tide_to_plipoints(tidemodel=tidemodel, gdf_points=gdf_points, 
                                                     component_list=component_list, load=True)
    ForcingModel_object = dfmt.plipointsDataset_to_ForcingModel(plipointsDataset=data_interp)
    
    forcing0 = ForcingModel_object.forcing[0]
    assert isinstance(ForcingModel_object, hcdfm.ForcingModel)
    assert isinstance(forcing0, hcdfm.Astronomic)
    
    assert forcing0.quantityunitpair[0].unit == '-'
    assert forcing0.quantityunitpair[0].quantity == 'astronomic component'
    assert forcing0.quantityunitpair[1].unit == 'm'
    assert forcing0.quantityunitpair[1].quantity == 'waterlevelbnd amplitude'
    assert forcing0.quantityunitpair[2].unit == 'degrees'
    assert forcing0.quantityunitpair[2].quantity == 'waterlevelbnd phase'


@pytest.mark.unittest
def test_interp_uds_to_plipoints():
    """
    very basic test for function, 
    should be made more strict with learnings from postprocess_interpolate_uds_toplipoints.py
    """
    
    ncbnd_construct = get_ncbnd_construct()
    varn_pointname = ncbnd_construct['varn_pointname']
    
    file_nc = dfmt.data.fm_grevelingen_map(return_filepath=True)
    uds = dfmt.open_partitioned_dataset(file_nc)
    
    # adding xy=[1,1] deliberately to test if there are nans included
    gdf = gpd.GeoDataFrame(geometry=gpd.points_from_xy([51500,55000,1],[418800,421500,1]))
    gdf[varn_pointname] = ['pt_0001','pt_0002','pt_0003']
    
    # fig, ax = plt.subplots()
    # uds.mesh2d_flowelem_bl.ugrid.plot(ax=ax)
    # gdf.plot(ax=ax)
    
    # interpolate depths (zsigma to z):
    ds_atdepths = dfmt.get_Dataset_atdepths(data_xr=uds, depths=[-5,-1])
    ds_atdepths = ds_atdepths.rename({'depth_from_z0':'depth'})
    
    #interpolate to plipoints
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        ds_plipoints = dfmt.interp_uds_to_plipoints(uds=ds_atdepths, gdf=gdf) #workaround for plipoints out of the model domain
    assert varn_pointname in ds_plipoints.coords
    assert 'depth' in ds_plipoints.coords # TODO: maybe should be z like ncbnd_construct['varn_depth'], although get_Dataset_atdepths() makes depth variable
    
    retrieved = ds_plipoints.mesh2d_sa1.isel(time=-1).to_numpy()
    expected = np.array([[29.00111981, 28.99379263],
                         [28.95170139, 28.63716083],
                         [np.nan, np.nan]])
    
    assert np.allclose(retrieved, expected, equal_nan=True)


@pytest.mark.unittest
def test_ext_add_boundary_object_per_polyline_onepolyline(tmp_path):
    file_pli = os.path.join(tmp_path,'test_model.pli')
    with open(file_pli,'w') as f:
        f.write("""name1
        2    2
        1.0    2.0
        3.0    4.0
        """)
    
    # new ext
    ext_new = hcdfm.ExtModel()
    forcingmodel_object = hcdfm.ForcingModel()
    boundary_object = hcdfm.Boundary(quantity='waterlevelbnd',
                                     locationfile=file_pli,
                                     forcingfile=forcingmodel_object)
    ext_add_boundary_object_per_polyline(ext_new=ext_new, boundary_object=boundary_object)
    assert len(ext_new.boundary) == 1
    polyfile_names = [os.path.basename(x.locationfile.filepath) for x in ext_new.boundary]
    assert polyfile_names == ['test_model.pli']


@pytest.mark.unittest
def test_ext_add_boundary_object_per_polyline_twopolylines(tmp_path):
    file_pli = os.path.join(tmp_path,'test_model.pli')
    with open(file_pli,'w') as f:
        f.write("""name1
        2    2
        1.0    2.0
        3.0    4.0
        name2
        2    2
        1.0    2.0
        3.0    4.0
        """)
    
    # new ext
    ext_new = hcdfm.ExtModel()
    forcingmodel_object = hcdfm.ForcingModel()
    boundary_object = hcdfm.Boundary(quantity='waterlevelbnd',
                                     locationfile=file_pli,
                                     forcingfile=forcingmodel_object)
    ext_add_boundary_object_per_polyline(ext_new=ext_new, boundary_object=boundary_object)
    assert len(ext_new.boundary) == 2
    polyfile_names = [os.path.basename(x.locationfile.filepath) for x in ext_new.boundary]
    assert polyfile_names == ['name1.pli', 'name2.pli']


@pytest.mark.unittest
def test_ext_add_boundary_object_per_polyline_wrong_name(tmp_path):
    file_pli = os.path.join(tmp_path,'test_model.pli')
    with open(file_pli,'w') as f:
        f.write("""test_model
        2    2
        1.0    2.0
        3.0    4.0
        name2
        2    2
        1.0    2.0
        3.0    4.0
        """)
    
    # new ext
    ext_new = hcdfm.ExtModel()
    forcingmodel_object = hcdfm.ForcingModel()
    boundary_object = hcdfm.Boundary(quantity='waterlevelbnd',
                                     locationfile=file_pli,
                                     forcingfile=forcingmodel_object)
    with pytest.raises(ValueError) as e:
        ext_add_boundary_object_per_polyline(ext_new=ext_new, boundary_object=boundary_object)
    assert "The names of one of the polylines in the polyfile is the same as the polyfilename" in str(e.value)


@pytest.mark.unittest
def test_open_dataset_extra_deprecated():
    with pytest.raises(DeprecationWarning):
        dfmt.open_dataset_extra()


@pytest.mark.unittest
def test_interp_hisnc_to_plipoints(tmp_path):
    ncbnd_construct = get_ncbnd_construct()
    dimn_point = ncbnd_construct['dimn_point']
    
    file_pli = os.path.join(tmp_path,'grev_internal.pli')
    with open(file_pli,'w') as f:
        f.write("""grev_internal
        4    2
        51655.45  420059.29
        55133.26  421410.41
        59848.92  422508.19
        61676.24  417399.27
        """)

    ds_his = dfmt.data.fm_grevelingen_his()
    da = ds_his[['waterlevel']]

    ds_interp = dfmt.interp_hisnc_to_plipoints(data_xr_his=da, file_pli=file_pli, kdtree_k=4)
    assert ds_interp.sizes[dimn_point] == 4
    assert ds_interp.sizes["time"] == 145

    last_points = ds_interp.waterlevel.isel(time=-1).to_numpy()
    last_points_expected = np.array(
        [-0.20098736, -0.19833495, -0.19681664, -0.19467336]
        )
    assert np.allclose(last_points, last_points_expected)
    pointnames_np = ds_interp.station_id.to_numpy()
    pointnames = [str(x) for x in pointnames_np]
    pointnames_expected = [
        'grev_internal_0001',
        'grev_internal_0002',
        'grev_internal_0003',
        'grev_internal_0004',
        ]
    assert pointnames == pointnames_expected
    
    assert set(ds_interp.coords) == set(['lat', 'lon', 'station_id', 'time'])



import numpy as np
import datetime as dt
import re
import xugrid as xu
import xarray as xr
import matplotlib.pyplot as plt
from dfm_tools.xarray_helpers import Dataset_varswithdim
from dfm_tools.xugrid_helpers import get_vertical_dimensions, decode_default_fillvals

__all__ = ["polyline_mapslice",
           "reconstruct_zw_zcc",
           "get_Dataset_atdepths",
           "rasterize_ugrid",
           "plot_ztdata",
    ]

def calc_dist_pythagoras(x1,x2,y1,y2):
    distance = np.sqrt((x2 - x1)**2 + (y2 - y1)**2)
    return distance


def calc_dist_haversine(lon1,lon2,lat1,lat2):
    #calculates distance between lat/lon coordinates in meters
    #https://community.esri.com/t5/coordinate-reference-systems-blog/distance-on-a-sphere-the-haversine-formula/ba-p/902128
    
    # convert to radians
    lon1_rad = np.deg2rad(lon1)
    lon2_rad = np.deg2rad(lon2)
    lat1_rad = np.deg2rad(lat1)
    lat2_rad = np.deg2rad(lat2)
    
    # apply formulae
    a = np.sin((lat2_rad-lat1_rad)/2)**2 + np.cos(lat1_rad) * np.cos(lat2_rad) * np.sin((lon2_rad-lon1_rad)/2)**2
    c = 2 * np.arctan2( np.sqrt(a), np.sqrt(1-a) )
    R = 6371000
    distance = R * c
    if np.isnan(distance).any():
        raise ValueError('nan encountered in calc_dist_latlon distance')
    return distance


def intersect_edges_withsort(uds,edges): #TODO: move sorting to xugrid? https://deltares.github.io/xugrid/api/xugrid.Ugrid2d.intersect_edges.html
    
    edge_index, face_index, intersections = uds.grid.intersect_edges(edges) #TODO: is fast, but maybe speed can be increased with bounding box?
    
    #ordering of face_index is wrong (visible with cb3 with long line_array), so sort on distance from startpoint (in x/y units)
    
    #compute distance from start of line to start of each linepart
    edge_len = np.linalg.norm(edges[:,1] - edges[:,0], axis=1)
    edge_len_cum = np.cumsum(edge_len)
    edge_len_cum0 = np.concatenate([[0],edge_len_cum[:-1]])
    
    #compute distance from start to lineparts to start of line (via line)
    startcoord_linepart = edges[edge_index,0,:]
    dist_tostart_linepart = np.linalg.norm(intersections[:,0,:] - startcoord_linepart, axis=1)
    dist_tostart_line = dist_tostart_linepart + edge_len_cum0[edge_index]
    
    #sorting on distance
    id_sorted = np.argsort(dist_tostart_line)
    edge_index = edge_index[id_sorted]
    face_index = face_index[id_sorted]
    intersections = intersections[id_sorted]
    return edge_index, face_index, intersections


def get_xzcoords_onintersection(uds, face_index, crs_dist_starts, crs_dist_stops):
    #TODO: remove hardcoding of variable names
    if 'time' in uds.dims: #TODO: maybe make time dependent grid?
        raise Exception('time dimension present in uds, provide uds.isel(time=timestep) instead. This is necessary to retrieve correct waterlevel or fullgrid output')
    
    dimn_layer, dimn_interfaces = get_vertical_dimensions(uds)
    gridname = uds.grid.name
    
    #construct fullgrid info (zcc/zw) for 3D models
    if dimn_layer in uds.dims:
        uds = reconstruct_zw_zcc(uds)

    #drop all variables that do not contain a face dimension, then select only all sliced faceidx
    xu_facedim = uds.grid.face_dimension
    face_index_xr = xr.DataArray(face_index,dims=('ncrossed_faces'))
    uds = Dataset_varswithdim(uds,dimname=xu_facedim) #TODO: is there an xugrid alternative?
    uds_sel = uds.sel({xu_facedim:face_index_xr})
    
    # take zvals_interface
    if dimn_layer in uds_sel.dims: #3D model
        nlay = uds.sizes[dimn_layer]
        zvals_interface_filled = uds_sel[f'{gridname}_flowelem_zw'].bfill(dim=dimn_interfaces) #fill nan values (below bed) with equal values
        zvals_interface = zvals_interface_filled.to_numpy().T #transpose to make in line with 2D sigma dataset
    else: #2D model, no layers
        nlay = 1
        data_frommap_wl3_sel = uds_sel[f'{gridname}_s1'].to_numpy() #TODO: add escape for missing wl/bl vars
        data_frommap_bl_sel = uds_sel[f'{gridname}_flowelem_bl'].to_numpy()
        zvals_interface = np.linspace(data_frommap_bl_sel,data_frommap_wl3_sel,nlay+1)

    #derive crs_verts
    crs_dist_starts_matrix = np.repeat(crs_dist_starts[np.newaxis],nlay,axis=0)
    crs_dist_stops_matrix = np.repeat(crs_dist_stops[np.newaxis],nlay,axis=0)
    crs_verts_x_all = np.array([[crs_dist_starts_matrix.ravel(),crs_dist_stops_matrix.ravel(),crs_dist_stops_matrix.ravel(),crs_dist_starts_matrix.ravel()]]).T
    crs_verts_z_all = np.ma.array([zvals_interface[1:,:].ravel(),zvals_interface[1:,:].ravel(),zvals_interface[:-1,:].ravel(),zvals_interface[:-1,:].ravel()]).T[:,:,np.newaxis]
    crs_verts = np.ma.concatenate([crs_verts_x_all, crs_verts_z_all], axis=2)
    
    #define grid
    shape_crs_grid = crs_verts[:,:,0].shape
    shape_crs_flat = crs_verts[:,:,0].ravel().shape
    xr_crs_grid = xu.Ugrid2d(node_x=crs_verts[:,:,0].ravel(),
                             node_y=crs_verts[:,:,1].ravel(),
                             fill_value=-1,
                             face_node_connectivity=np.arange(shape_crs_flat[0]).reshape(shape_crs_grid),
                             )

    #define dataset
    if dimn_layer in uds_sel.dims:
        crs_plotdata_clean = uds_sel.stack({xr_crs_grid.face_dimension:[dimn_layer,'ncrossed_faces']},create_index=False)
    else: #2D: still make sure xr_crs_grid.face_dimension is created, using stack since .rename() gives "UserWarning: rename 'ncrossed_faces' to 'mesh2d_nFaces' does not create an index anymore."
        crs_plotdata_clean = uds_sel.stack({xr_crs_grid.face_dimension:['ncrossed_faces']},create_index=False)
                    
    #combine into xugrid
    xr_crs_ugrid = xu.UgridDataset(crs_plotdata_clean, grids=[xr_crs_grid])
    return xr_crs_ugrid


def polyline_mapslice(uds:xu.UgridDataset, line_array:np.array) -> xu.UgridDataset:
    """
    Slice trough mapdata, combine: intersect_edges_withsort, calculation of distances and conversion to ugrid dataset.

    Parameters
    ----------
    uds : xu.UgridDataset
        DESCRIPTION.
    line_array : np.array
        DESCRIPTION.
    calcdist_fromlatlon : bool, optional
        DESCRIPTION. The default is None.

    Raises
    ------
    Exception
        DESCRIPTION.

    Returns
    -------
    xr_crs_ugrid : xu.UgridDataset
        DESCRIPTION.

    """
    
    #compute intersection coordinates of crossings between edges and faces and their respective indices
    edges = np.stack([line_array[:-1],line_array[1:]],axis=1)
    edge_index, face_index, intersections = intersect_edges_withsort(uds=uds, edges=edges)
    if len(edge_index) == 0:
        raise ValueError('polyline does not cross mapdata')
    
    if uds.grids[0].is_geographic:
        calc_dist = calc_dist_haversine
    else:
        calc_dist = calc_dist_pythagoras
    
    #compute pyt/haversine start/stop distances for all intersections
    edge_len = calc_dist(edges[:,0,0], edges[:,1,0], edges[:,0,1], edges[:,1,1])
    edge_len_cum = np.cumsum(edge_len)
    edge_len_cum0 = np.concatenate([[0],edge_len_cum[:-1]])
    crs_dist_starts = calc_dist(edges[edge_index,0,0], intersections[:,0,0], edges[edge_index,0,1], intersections[:,0,1]) + edge_len_cum0[edge_index]
    crs_dist_stops  = calc_dist(edges[edge_index,0,0], intersections[:,1,0], edges[edge_index,0,1], intersections[:,1,1]) + edge_len_cum0[edge_index]
    
    #derive vertices from cross section (distance from first point)
    xr_crs_ugrid = get_xzcoords_onintersection(uds=uds, face_index=face_index, crs_dist_starts=crs_dist_starts, crs_dist_stops=crs_dist_stops)
    
    return xr_crs_ugrid


def get_formula_terms(uds, varn_contains):
    """
    get formula_terms for zw/zcc reconstruction, convert to list and then to dict. This can be done for layer/interface (via varn_contains)
    """
    osz_varnames = list(uds.filter_by_attrs(formula_terms=lambda v: v is not None).variables) #names of variables containing attribute "formula_terms"
    osz_varnames_contains = [x for x in osz_varnames if varn_contains in x] #TODO: to get the layer/interface ocean_*_coordinate. Not too pretty, but it works
    if len(osz_varnames_contains) != 1: #should be 1 exactly, none is the case in zlayer models
        raise ValueError(f'no or more than one {varn_contains} variable found with formula_terms attribute: {osz_varnames}')
    osz_varn = osz_varnames_contains[0]
    osz_formulaterms = uds[osz_varn].attrs['formula_terms']
    tokens = re.split('[:\\s]+', osz_formulaterms)
    osz_formulaterms_dict = dict(zip(tokens[::2], tokens[1::2]))
    return osz_formulaterms_dict


def reconstruct_zw_zcc_fromsigma(uds):
    """
    reconstruct full grid output (time/face-varying z-values) for sigma model, necessary for slicing sigmamodel on depth value
    based on https://cfconventions.org/cf-conventions/cf-conventions.html#_ocean_sigma_coordinate
    """
    osz_formulaterms_int_dict = get_formula_terms(uds,varn_contains='interface')
    osz_formulaterms_lay_dict = get_formula_terms(uds,varn_contains='layer')
    gridname = uds.grid.name
    
    uds_eta = uds[osz_formulaterms_int_dict['eta']] #mesh2d_s1
    uds_depth = uds[osz_formulaterms_int_dict['depth']] #mesh2d_bldepth in new output, mesh2d_waterdepth in old output (see comment below)
    if uds_depth.attrs['standard_name'] == 'sea_floor_depth_below_sea_surface': # previously the waterdepth instead of negative bedlevel was coupled via the formula_terms in sigmamodels (was fixed in OSS 140982 / 29-3-2022)
        uds_depth = -uds[f'{gridname}_flowelem_bl'] # assuming this variable is available, which is not guaranteed
    uds_sigma_int = uds[osz_formulaterms_int_dict['sigma']] #mesh2d_interface_sigma
    uds_sigma_lay = uds[osz_formulaterms_lay_dict['sigma']] #mesh2d_layer_sigma
    
    uds[f'{gridname}_flowelem_zw'] = uds_eta + uds_sigma_int*(uds_depth+uds_eta)
    uds[f'{gridname}_flowelem_zcc'] = uds_eta + uds_sigma_lay*(uds_depth+uds_eta)
    
    uds = uds.set_coords([f'{gridname}_flowelem_zw',f'{gridname}_flowelem_zcc'])
    return uds


def reconstruct_zw_zcc_fromz(uds):
    """
    reconstruct full grid output (time/face-varying z-values) for zvalue model. Necessary when extracting values with zdepth w.r.t. waterlevel/bedlevel
    """
    
    dimn_layer, dimn_interfaces = get_vertical_dimensions(uds)
    gridname = uds.grid.name
    
    uds_eta = uds[f'{gridname}_s1'] # assuming this variable is available, which is not guaranteed
    uds_z0 = xu.zeros_like(uds_eta)
    uds_bl = uds[f'{gridname}_flowelem_bl'] # assuming this variable is available, which is not guaranteed
    
    #deriving zcenter values, clipping zcc to bl/wl
    # zvals_center = uds[f'{gridname}_layer_z']
    # zcc = (uds_z0+zvals_center).clip(min=uds_bl, max=uds_eta)

    #deriving zinterface values, first expanding zint to wl.max(), then clipping zw to bl/wl
    zvals_interface = uds[f'{gridname}_interface_z']
    # make sure mesh2d_interface_z.max()>=wl.max() (is clipped to wl again in next step)
    if zvals_interface[-1] < uds_eta.max():
        zvals_interface[-1] = uds_eta.max()
    zw = (uds_z0+zvals_interface).clip(min=uds_bl, max=uds_eta)
       
    # correction: set interfaces below bed to nan (keeping the interface at the bed with shift)
    bool_belowbed = zw.shift({dimn_interfaces:-1}) <= uds_bl
    zw = zw.where(~bool_belowbed)
    # correction: set interfaces above wl to nan (keeping the interface at the wl with shift)
    bool_abovewl = zw.shift({dimn_interfaces:1}) >= uds_eta
    zw = zw.where(~bool_abovewl)
    
    # correction: set centers to values in between interfaces (and reshape+rename from int to cent dim)
    zcc = zw.rolling({dimn_interfaces:2}).mean()
    zcc = zcc.isel({dimn_interfaces:slice(1,None)}).rename({dimn_interfaces:dimn_layer})
    
    uds[f'{gridname}_flowelem_zw'] = zw
    uds[f'{gridname}_flowelem_zcc'] = zcc
    uds = uds.set_coords([f'{gridname}_flowelem_zw',f'{gridname}_flowelem_zcc'])
    return uds


def reconstruct_zw_zcc_fromzsigma(uds):
    """
    Reconstruct full grid output (time/face-varying z-values) for zsigmavalue model
    without full grid output. Implemented in FM kernel with
    https://issuetracker.deltares.nl/browse/UNST-5477. The reconstruction is based on
    https://cfconventions.org/cf-conventions/cf-conventions.html#_ocean_sigma_over_z_coordinate
    """
    
    dimn_layer, dimn_interfaces = get_vertical_dimensions(uds)
    gridname = uds.grid.name
    
    # TODO: temporarily decode default fillvalues, xarray only decodes explicitly set
    # fillvalues: https://github.com/Deltares/dfm_tools/issues/490
    uds = xu.UgridDataset(decode_default_fillvals(uds.obj), grids=uds.grids)
    
    osz_formulaterms_int_dict = get_formula_terms(uds,varn_contains='interface')
    # osz_formulaterms_lay_dict = get_formula_terms(uds,varn_contains='layer')
    
    uds_eta = uds[osz_formulaterms_int_dict['eta']] #mesh2d_s1
    # mesh2d_bldepth positive version of mesh2d_flowelem_bl (but always present)
    uds_depth = uds[osz_formulaterms_int_dict['depth']] #mesh2d_bldepth
    uds_depth_c = uds[osz_formulaterms_int_dict['depth_c']] #mesh2d_sigmazdepth
    uds_zlev_int = uds[osz_formulaterms_int_dict['zlev']] #mesh2d_interface_z
    uds_sigma_int = uds[osz_formulaterms_int_dict['sigma']] #mesh2d_interface_sigma
    # uds_zlev_lay = uds[osz_formulaterms_lay_dict['zlev']] #mesh2d_layer_z
    # uds_sigma_lay = uds[osz_formulaterms_lay_dict['sigma']] #mesh2d_layer_sigma
    
    # for levels k where sigma(k) has a defined value and zlev(k) is not defined:
    # z(n,k,j,i) = eta(n,j,i) + sigma(k)*(min(depth_c,depth(j,i))+eta(n,j,i))
    zw_sigmapart = uds_eta + uds_sigma_int*(uds_depth.clip(max=uds_depth_c) + uds_eta)
    # zcc_sigmapart = uds_eta + uds_sigma_lay*(uds_depth.clip(max=uds_depth_c)+uds_eta)
    # for levels k where zlev(k) has a defined value and sigma(k) is not defined: 
    # z(n,k,j,i) = zlev(k)
    # clip zvalues with bedlevel and waterlevel
    zw_zpart = uds_zlev_int.clip(min=-uds_depth, max=uds_eta)
    # zcc_zpart = uds_zlev_lay.clip(min=-uds_depth)
    zw = zw_sigmapart.fillna(zw_zpart)
    # zcc = zcc_sigmapart.fillna(zcc_zpart)
    
    # correction: set interfaces below bed to nan (shift preserves interface at the bed)
    bool_belowbed = zw.shift({dimn_interfaces:-1}) <= -uds_depth
    zw = zw.where(~bool_belowbed)
    # correction: set interfaces above wl to nan (keeping the interface at the wl with shift)
    # required for waterlevels < zsigma-interface: https://github.com/Deltares/dfm_tools/issues/1218
    bool_abovewl = zw.shift({dimn_interfaces:1}) >= uds_eta
    zw = zw.where(~bool_abovewl)

    # correction: set centers to values in between interfaces
    # and reshape+rename from interfaces to centers dimension
    zcc = zw.rolling({dimn_interfaces:2}).mean()
    zcc = zcc.isel({dimn_interfaces:slice(1,None)}).rename({dimn_interfaces:dimn_layer})
    
    uds[f'{gridname}_flowelem_zw'] = zw
    uds[f'{gridname}_flowelem_zcc'] = zcc
    uds = uds.set_coords([f'{gridname}_flowelem_zw',f'{gridname}_flowelem_zcc'])
    return uds


def reconstruct_zw_zcc(uds:xu.UgridDataset):
    """
    reconstruct full grid output (time/face-varying z-values) for all layertypes,
    calls the respective reconstruction function

    Parameters
    ----------
    uds : xu.UgridDataset
        DESCRIPTION.

    Raises
    ------
    KeyError
        DESCRIPTION.

    Returns
    -------
    uds : TYPE
        DESCRIPTION.

    """
    
    gridname = uds.grid.name
    varname_zint = f'{gridname}_flowelem_zw'
    
    #reconstruct zw/zcc variables (if not in file) and treat as fullgrid mapfile from here
    if varname_zint in uds.variables: #fullgrid info already available, so continuing
        print(f'zw/zcc (fullgrid) values already present in Dataset in variable {varname_zint}')
    elif len(uds.filter_by_attrs(standard_name='ocean_sigma_z_coordinate')) != 0:
        print('zsigma-layer model, computing zw/zcc (fullgrid) values and treat as fullgrid model from here')
        uds = reconstruct_zw_zcc_fromzsigma(uds)
    elif len(uds.reset_coords().filter_by_attrs(standard_name='ocean_sigma_coordinate')) != 0:
        print('sigma-layer model, computing zw/zcc (fullgrid) values and treat as fullgrid model from here')
        uds = reconstruct_zw_zcc_fromsigma(uds)
    elif f'{gridname}_layer_z' in uds.variables:
        print('z-layer model, computing zw/zcc (fullgrid) values and treat as fullgrid model from here')
        uds = reconstruct_zw_zcc_fromz(uds)
    else:
        raise KeyError(f'layers present, but unknown layertype, expected one of variables: {gridname}_flowelem_zw, {gridname}_layer_sigma, {gridname}_layer_z')
    return uds

    
def get_Dataset_atdepths(data_xr:xu.UgridDataset, depths, reference:str ='z0'):    
    """
    Lazily depth-slice a dataset with layers. Performance can be increased by using a subset of variables or subsetting the dataset in any dimension.
    This can be done for instance with ds.isel(time=-1) or uds.ugrid.sel(x=slice(),y=slice()) to subset a ugrid dataset in space.
    The return dataset only contains the sliced variables.
    
    Parameters
    ----------
    data_xr : xu.UgridDataset
        has to be Dataset (not a DataArray), otherwise mesh2d_flowelem_zw etc are not available (interface z values)
        in case of zsigma/sigma layers (or fullgrid), it is advisable to .sel()/.isel() the time dimension first, because that is less computationally heavy
    depths : TYPE
        int/float or list/array of int/float. Depths w.r.t. reference level. If reference=='waterlevel', depth>0 returns only nans. If reference=='bedlevel', depth<0 returns only nans. Depths are sorted and only uniques are kept.
    reference : str, optional
        compute depth w.r.t. z0/waterlevel/bed. The default is 'z0'.
    zlayer_z0_selnearest : bool, optional
        Use xr.interp() to interpolate zlayer model to z-value. Only possible for reference='z' (not 'waterlevel' or 'bedlevel'). Only used if "mesh2d_layer_z" is present (zlayer model)
        This is faster but results in values interpolated between zcc (z cell centers), so it is different than slicing.. The default is False.

    Raises
    ------
    Exception
        DESCRIPTION.

    Returns
    -------
    xu.UgridDataset
        Dataset with the depth-sliced variables.

    """
    
    depth_vardimname = f'depth_from_{reference}'
    
    dimn_layer, dimn_interfaces = get_vertical_dimensions(data_xr)
    
    if dimn_layer is not None: #D-FlowFM mapfile
        gridname = data_xr.grid.name
        varname_zint = f'{gridname}_flowelem_zw'
        dimname_layc = dimn_layer
        dimname_layw = dimn_interfaces
        varname_wl = f'{gridname}_s1'
        varname_bl = f'{gridname}_flowelem_bl'
    elif 'laydim' in data_xr.dims: #D-FlowFM hisfile
        varname_zint = 'zcoordinate_w'
        dimname_layc = 'laydim'
        dimname_layw = 'laydimw'
        varname_wl = 'waterlevel'
        varname_bl = 'bedlevel'
    else:
        print(UserWarning('depth/layer dimension not found, probably 2D model, returning input Dataset')) #TODO: this can also be at depth, since slice will put parts of model dry (and allnan if below wl or below bl). Implement this
        return data_xr #early return
    
    #create depth xr.DataArray
    if isinstance(depths,(float,int)):
        depth_dims = ()
    else:
        depths = np.unique(depths) #array of unique+sorted floats/ints
        depth_dims = (depth_vardimname)
    depths_xr = xr.DataArray(depths,dims=depth_dims,attrs={'units':'m',
                                                           'reference':f'model_{reference}',
                                                           'positive':'up'}) #TODO: make more in line with CMEMS etc
    
    #potentially construct fullgrid info (zcc/zw)
    if dimn_layer is not None: #D-FlowFM mapfile
        data_xr = reconstruct_zw_zcc(data_xr)
    
    #correct reference level
    if reference=='z0':
        zw_reference = data_xr[varname_zint]
    elif reference=='waterlevel':
        if varname_wl not in data_xr.variables:
            raise KeyError(f'get_Dataset_atdepths() called with reference=waterlevel, but {varname_wl} variable not present')
        data_wl = data_xr[varname_wl]
        zw_reference = data_xr[varname_zint] - data_wl
    elif reference=='bedlevel':
        if varname_bl not in data_xr.variables:
            raise KeyError(f'get_Dataset_atdepths() called with reference=bedlevel, but {varname_bl} variable not present') #TODO: in case of zsigma/sigma it can also be -mesh2d_bldepth
        data_bl = data_xr[varname_bl]
        zw_reference = data_xr[varname_zint] - data_bl
    else:
        raise KeyError(f'unknown reference "{reference}" (options are z0, waterlevel and bedlevel)')
    
    print('>> subsetting data on fixed depth in fullgrid z-data: ',end='')
    dtstart = dt.datetime.now()
    
    #get layerbool via z-interface value (zw), check which celltop-interfaces are above/on depth and which which cellbottom-interfaces are below/on depth
    bool_topinterface_abovedepth = zw_reference.isel({dimname_layw:slice(1,None)}) >= depths_xr
    bool_botinterface_belowdepth = zw_reference.isel({dimname_layw:slice(None,-1)}) <= depths_xr
    bool_topbotinterface_arounddepth = bool_topinterface_abovedepth & bool_botinterface_belowdepth #this bool also automatically excludes all values below bed and above wl
    bool_topbotinterface_arounddepth = bool_topbotinterface_arounddepth.rename({dimname_layw:dimname_layc}) #correct dimname for interfaces to centers
    
    #subset variables that have no, time, face and/or layer dims, slice only variables with all three dims (and add to subset)
    bool_dims = [x for x in bool_topbotinterface_arounddepth.dims if x!=depth_vardimname] #exclude depth_vardimname (present if multiple depths supplied), since it is not present in pre-slice variables
    variables_toslice = [var for var in data_xr.data_vars if set(bool_dims).issubset(data_xr[var].dims)]
    
    #actual slicing with .where().max()
    ds_atdepths = data_xr[variables_toslice].where(bool_topbotinterface_arounddepth).max(dim=dimname_layc,keep_attrs=True) #set all layers but one to nan, followed by an arbitrary reduce (max in this case) #TODO: check if attributes should be passed/altered
    #TODO: suppress warning (upon plotting/load/etc): "C:\Users\veenstra\Anaconda3\envs\dfm_tools_env\lib\site-packages\dask\array\reductions.py:640: RuntimeWarning: All-NaN slice encountered" >> already merged in xarray: https://github.com/dask/dask/pull/9916
    ds_atdepths = ds_atdepths.drop_dims([dimname_layw,dimname_layc],errors='ignore') #dropping interface dim if it exists, since it does not correspond to new depths dim
    
    #add depth as coordinate var
    ds_atdepths[depth_vardimname] = depths_xr
    ds_atdepths = ds_atdepths.set_coords([depth_vardimname])
    
    print(f'{(dt.datetime.now()-dtstart).total_seconds():.2f} sec')
    
    return ds_atdepths


def rasterize_ugrid(uds:xu.UgridDataset, ds_like:xr.Dataset = None, resolution:float = None):
    """
    Rasterizing ugrid dataset to regular dataset. ds_like has higher priority than `resolution`. If both are not passed, a raster is generated of at least 200x200
    inspired by xugrid.plot.imshow and xugrid.ugrid.ugrid2d.rasterize/rasterize_like.
    
    Parameters
    ----------
    uds : xu.UgridDataset
        DESCRIPTION.
    ds_like : xr.Dataset, optional
        xr.Dataset with ed x and y variables to interpolate uds to. The default is None.
    resolution : float, optional
        Only used if ds_like is not supplied. The default is None.
    
    Raises
    ------
    Exception
        DESCRIPTION.

    Returns
    -------
    ds : TYPE
        DESCRIPTION.

    """
    
    if isinstance(uds,xu.core.wrap.UgridDataset):
        xu_facedim = uds.grid.face_dimension
        uds_facevars = Dataset_varswithdim(uds,xu_facedim)
        face_str = f'Dataset with {len(uds_facevars.data_vars)} face variables'
    elif isinstance(uds,xu.core.wrap.UgridDataArray):
        face_str = 'DataArray'
    else:
        raise TypeError(f'rasterize_ugrid expected xu.core.wrap.UgridDataset or xu.core.wrap.UgridDataArray, got {type(uds)} instead')
    
    if ds_like is None:
        xmin, ymin, xmax, ymax = uds.grid.bounds
        dx = xmax - xmin
        dy = ymax - ymin
        if resolution is None: # check if a rasterization resolution is passed, otherwise default to 200 raster cells otherwise for the smallest axis.
            resolution = min(dx, dy) / 200
        d = abs(resolution)
        regx = np.arange(xmin + 0.5 * d, xmax, d)
        regy = np.arange(ymin + 0.5 * d, ymax, d)
        ds_like = xr.DataArray(np.empty((len(regy), len(regx))), {"y": regy, "x": regx}, ["y", "x"])
    
    print(f'>> rasterizing ugrid {face_str} to shape=({len(ds_like.y)},{len(ds_like.x)}): ',end='')
    dtstart = dt.datetime.now()
    ds = uds.ugrid.rasterize_like(other=ds_like)
    print(f'{(dt.datetime.now()-dtstart).total_seconds():.2f} sec')
    
    return ds


def plot_ztdata(data_xr_sel, varname, ax=None, only_contour=False, **kwargs):
    """
    

    Parameters
    ----------
    data_xr : TYPE
        DESCRIPTION.
    varname : TYPE
        DESCRIPTION.
    ax : matplotlib.axes._subplots.AxesSubplot, optional
        the figure axis. The default is None.
    only_contour : bool, optional
        Wheter to plot contour lines of the dataset. The default is False.
    kwargs : TYPE
        properties to give on to the pcolormesh function.
    
    Raises
    ------
    Exception
        DESCRIPTION.

    Returns
    -------
    pc : matplotlib.collections.QuadMesh
        DESCRIPTION.
    
    """
    
    if not ax: ax=plt.gca()
    
    if len(data_xr_sel[varname].shape) != 2:
        raise ValueError(f'unexpected number of dimensions in requested squeezed variable ({data_xr_sel[varname].shape}), first use data_xr.isel(stations=int) to select a single station') #TODO: can also have a different cause, improve message/testing?
    
    #repair zvalues at wl/wl (filling nans and clipping to wl/bl). bfill replaces nan values with last valid value, this is necessary to enable pcolormesh to work. clip forces data to be within bl/wl
    #TODO: put clip in preproces_hisnc to make plotting easier?
    data_xr_sel['zcoordinate_c'] = data_xr_sel['zcoordinate_c'].bfill(dim='laydim').clip(min=data_xr_sel.bedlevel,max=data_xr_sel.waterlevel)
    data_xr_sel['zcoordinate_w'] = data_xr_sel['zcoordinate_w'].bfill(dim='laydimw').clip(min=data_xr_sel.bedlevel,max=data_xr_sel.waterlevel)
    
    # generate 2 2d grids for the x & y bounds (you can also give one 2D array as input in case of eg time varying z coordinates)
    data_fromhis_zcor = data_xr_sel['zcoordinate_w'].to_numpy() 
    data_fromhis_zcor = np.concatenate([data_fromhis_zcor,data_fromhis_zcor[[-1],:]],axis=0)
    time_np = data_xr_sel.time.to_numpy()
    time_cor = np.concatenate([time_np,time_np[[-1]]])
    time_mesh_cor = np.tile(time_cor,(data_fromhis_zcor.shape[-1],1)).T
    if only_contour:
        pc = data_xr_sel[varname].plot.contour(ax=ax, x='time', y='zcoordinate_c', **kwargs)
    else:
        #pc = data_xr_sel[varname].plot.pcolormesh(ax=ax, x='time', y='zcoordinate_w', **kwargs) #TODO: not possible to put center values on interfaces, so more difficult approach needed
        pc = ax.pcolormesh(time_mesh_cor, data_fromhis_zcor, data_xr_sel[varname], **kwargs)
   
    return pc




# -*- coding: utf-8 -*-
"""
Created on Thu Oct 26 16:08:26 2023

@author: veenstra
"""

import os
import pytest
import dfm_tools as dfmt
import pandas as pd
import hydrolib.core.dflowfm as hcdfm
import xarray as xr
import numpy as np
from dfm_tools.modelbuilder import get_quantity_list, get_ncvarname


@pytest.mark.unittest
def test_get_quantity_list():
    quantity_list = get_quantity_list('uxuyadvectionvelocitybnd')
    assert quantity_list == ['ux','uy']
    quantity_list = get_quantity_list('salinitybnd')
    assert quantity_list == ['salinitybnd']
    quantity_list = get_quantity_list(['salinitybnd'])
    assert quantity_list == ['salinitybnd']


@pytest.mark.unittest
def test_get_ncvarname_list():
    conversion_dict = dfmt.get_conversion_dict()
    
    ncvarname = get_ncvarname(quantity='salinitybnd', conversion_dict=conversion_dict)
    assert ncvarname == "so"
    
    with pytest.raises(KeyError) as e:
        get_ncvarname(quantity='nonexistingbnd', conversion_dict=conversion_dict)
    assert "quantity 'nonexistingbnd' not in conversion_dict" in str(e.value)


@pytest.mark.unittest
def test_constant_to_bc(tmp_path):
    # generate pli with two polylines
    file_pli = tmp_path / "test.pli"
    with open(file_pli,'w') as f:
        f.write("""name1
        2    2
        1.0    2.0
        3.0    4.0
        name2
        2    2
        1.0    2.0
        3.0    4.0
        """)
    # generate ext and add constant forcing
    ext_new = hcdfm.ExtModel()
    _ = dfmt.constant_to_bc(ext_new=ext_new, file_pli=file_pli, constant=0.5)
    # check file existence
    file_bc = tmp_path / "waterlevelbnd_constant_test.bc"
    assert os.path.exists(file_bc)
    # check contents
    forcing_obj = hcdfm.ForcingModel(file_bc)
    assert len(forcing_obj.forcing) == 2
    assert np.allclose(forcing_obj.forcing[0].datablock, [[0.5]])
    assert np.allclose(forcing_obj.forcing[1].datablock, [[0.5]])


@pytest.mark.parametrize("timecase", [pytest.param(x, id=x) for x in ['midnight','noon','monthly']])
@pytest.mark.systemtest
def test_cmems_nc_to_bc(tmp_path, timecase, cmems_dataset_4times):
    """
    tests for midnight-centered data, noon-centered data and monthly timestamped data
    """
    file_pli = os.path.join(tmp_path,'test_model.pli')
    with open(file_pli,'w') as f:
        f.write("""name
                    2    2
                    -9.6   42.9
                    -9.5   43.0
                    """)
    
    ds = cmems_dataset_4times
    if timecase == "midnight":
        ds["time"] = ds["time"] + pd.Timedelta(hours=12)
    elif timecase == "monthly":
        ds["time"] = [pd.Timestamp("2019-11-01"),
                      pd.Timestamp("2019-12-01"),
                      pd.Timestamp("2020-01-01"),
                      pd.Timestamp("2020-02-01")]
    
    dir_pattern = os.path.join(tmp_path, "temp_cmems_4day_so.nc")
    file_nc = dir_pattern
    ds.to_netcdf(file_nc)
    
    ext_new = hcdfm.ExtModel()
    
    ext_new = dfmt.cmems_nc_to_bc(ext_new=ext_new,
                                  list_quantities=["salinitybnd"],
                                  tstart="2020-01-01",
                                  tstop="2020-01-03",
                                  file_pli=file_pli,
                                  dir_pattern=dir_pattern,
                                  dir_output=tmp_path,
                                  )
    
    file_expected = tmp_path / "salinitybnd_CMEMS_test_model.bc"
        
    if timecase == "midnight":
        times_expected =  [
            '2020-01-01 00:00:00',
            '2020-01-02 00:00:00',
            '2020-01-03 00:00:00',
            ]
    elif timecase == "noon":
        times_expected =  [
            '2019-12-31 12:00:00',
            '2020-01-01 12:00:00',
            '2020-01-02 12:00:00',
            '2020-01-03 12:00:00',
            ]
    elif timecase == "monthly":
        times_expected =  ['2020-01-01 00:00:00', '2020-02-01 00:00:00']
    
    assert os.path.exists(file_expected)
    forcing_obj = hcdfm.ForcingModel(file_expected)
    ds_out = dfmt.forcinglike_to_Dataset(forcing_obj.forcing[0])
    actual_times = ds_out.time.to_pandas().dt.strftime("%Y-%m-%d %H:%M:%S").tolist()
    assert actual_times == times_expected
    
    assert "salinitybnd" in ds_out.data_vars
    assert ds_out.salinitybnd.isnull().sum().load() == 0
    
    # check whether the cmems depth definition comes trough
    assert 'z' in ds_out.coords
    depths_expected = np.array([-0.494025, -1.541375, -2.645669, -3.819495, -5.078224])
    assert np.allclose(ds_out['z'].to_numpy(), depths_expected)
    assert ds_out['z'].attrs['positive'] == 'up'


@pytest.mark.parametrize("timecase", [pytest.param(x, id=x) for x in ['midnight','noon','monthly']])
@pytest.mark.systemtest
def test_cmems_nc_to_ini(tmp_path, timecase, cmems_dataset_4times):
    """
    tests for midnight-centered data, noon-centered data and monthly timestamped data
    """
    ds1 = cmems_dataset_4times.isel(time=slice(None,2))
    ds2 = cmems_dataset_4times.isel(time=slice(2,None))
    if timecase == "midnight":
        ds1["time"] = ds1["time"] + pd.Timedelta(hours=12)
        ds2["time"] = ds2["time"] + pd.Timedelta(hours=12)
    elif timecase == "monthly":
        ds1["time"] = [pd.Timestamp("2019-11-01"), pd.Timestamp("2019-12-01")]
        ds2["time"] = [pd.Timestamp("2020-01-01"), pd.Timestamp("2020-02-01")]
    
    dir_pattern = os.path.join(tmp_path, "temp_cmems_2day_*.nc")
    file_nc1 = dir_pattern.replace("*","so_p1")
    file_nc2 = dir_pattern.replace("*","so_p2")
    file_nc3 = dir_pattern.replace("*","thetao_p1")
    file_nc4 = dir_pattern.replace("*","thetao_p2")
    ds1.to_netcdf(file_nc1)
    ds2.to_netcdf(file_nc2)
    ds1.rename({"so":"thetao"}).to_netcdf(file_nc3)
    ds2.rename({"so":"thetao"}).to_netcdf(file_nc4)
    
    ext_old = hcdfm.ExtOldModel()
    
    dir_pattern_ncvarname = dir_pattern.replace("*","{ncvarname}*")
    
    ext_old = dfmt.cmems_nc_to_ini(ext_old=ext_old,
                                   dir_output=tmp_path,
                                   list_quantities=["salinitybnd"],
                                   tstart="2020-01-01",
                                   dir_pattern=dir_pattern_ncvarname)
    
    file_expected = tmp_path / "nudge_salinity_temperature_2020-01-01_00-00-00.nc"
    
    if timecase == "midnight":
        times_expected =  ['2020-01-01 00:00:00', '2020-01-02 00:00:00']
    elif timecase == "noon":
        times_expected =  ['2019-12-31 12:00:00', '2020-01-01 12:00:00']
    elif timecase == "monthly":
        times_expected =  ['2020-01-01 00:00:00', '2020-02-01 00:00:00']
        
    assert os.path.exists(file_expected)
    ds_out = xr.open_dataset(file_expected)
    
    times_actual = ds_out.time.to_pandas().dt.strftime("%Y-%m-%d %H:%M:%S").tolist()
    assert times_expected == times_actual
    
    assert "so" in ds_out.data_vars
    assert "thetao" in ds_out.data_vars
    assert ds_out.so.isnull().sum().load() == 0
    
    # check whether the cmems depth definition comes trough
    assert 'depth' in ds_out.coords
    depths_expected = np.array([-0.494025, -1.541375, -2.645669, -3.819495, -5.078224])
    assert np.allclose(ds_out['depth'].to_numpy(), depths_expected)
    assert ds_out['depth'].attrs['positive'] == 'up'


@pytest.mark.systemtest
def test_cmems_nc_to_ini_tracer(tmp_path, cmems_dataset_4times):
    """
    coverage for initialtracer
    """
    ds = cmems_dataset_4times
    ds = ds.rename({"so":"no3"})
    file_nc = os.path.join(tmp_path, "temp_cmems_4day_no3.nc")
    ds.to_netcdf(file_nc)
    
    ext_old = hcdfm.ExtOldModel()
    
    ext_old = dfmt.cmems_nc_to_ini(ext_old=ext_old,
                                   dir_output=tmp_path,
                                   list_quantities=["tracerbndNO3"],
                                   tstart="2020-01-01",
                                   dir_pattern=file_nc)
    
    file_expected = tmp_path / "initialtracerNO3_2020-01-01_00-00-00.nc"
    
    times_expected =  ['2019-12-31 12:00:00', '2020-01-01 12:00:00']
        
    assert os.path.exists(file_expected)
    ds_out = xr.open_dataset(file_expected)
    
    times_actual = ds_out.time.to_pandas().dt.strftime("%Y-%m-%d %H:%M:%S").tolist()
    assert times_expected == times_actual
    
    assert "initialtracerNO3" in ds_out.data_vars
    assert ds_out.initialtracerNO3.isnull().sum().load() == 0
    assert 'depth' in ds_out.coords


@pytest.mark.unittest
def test_cmems_nc_to_ini_skipped():
    list_quantities = ["temperaturebnd",
                       "ux",
                       "uxuyadvectionvelocitybnd",
                       "nonexisting",
                       ]
    
    _ = dfmt.cmems_nc_to_ini(
        ext_old=None, # dummy since it is not reached
        dir_output=None, # dummy since it is not reached
        list_quantities=list_quantities,
        tstart="2020-01-01",
        dir_pattern=None, # dummy since it is not reached
        )


@pytest.mark.unittest
def test_create_model_exec_files_none(tmp_path):
    mdu_file = tmp_path / "temp_test.mdu"
    file_dimr = tmp_path / "dimr_config.xml"
    
    nproc = 4 # number of processes
    dimrset_folder = None
    mdu = hcdfm.FMModel()
    mdu.save(mdu_file)
    dfmt.create_model_exec_files(file_mdu=mdu_file, nproc=nproc, dimrset_folder=dimrset_folder)
    
    assert os.path.isfile(file_dimr)
    
    # check if the dimr_config.xml file had the correct amount of processes
    # this failed before fixing https://github.com/Deltares/HYDROLIB-core/issues/562
    with open(file_dimr, 'r') as f:
        data = f.readlines()
    assert '    <process>0 1 2 3</process>\n' in data


@pytest.mark.unittest
def test_create_model_exec_files_docker(tmp_path):
    mdu_file = tmp_path / "temp_test.mdu"
    file_dimr = tmp_path / "dimr_config.xml"
    file_docker = tmp_path / "run_model.sh"
    
    nproc = 1 # number of processes
    dimrset_folder = "docker"
    mdu = hcdfm.FMModel()
    mdu.save(mdu_file)
    dfmt.create_model_exec_files(file_mdu=mdu_file, nproc=nproc, dimrset_folder=dimrset_folder)
    
    assert os.path.isfile(file_dimr)
    assert os.path.isfile(file_docker)
    
    # check for unix line endings, this is required to be found by docker somehow
    with open(file_docker, 'rb') as f:
        data = f.readline()
    assert data == b'#!/bin/bash\n'


@pytest.mark.unittest
def test_make_paths_relative(tmp_path):
    file_pli = os.path.join(tmp_path,'test_model.pli')
    with open(file_pli,'w') as f:
        f.write("""name
                    2    2
                    1.0    2.0
                    3.0    4.0
                    """)
    
    # new ext
    ext_file_new = os.path.join(tmp_path, 'test_new.ext')
    ext_new = hcdfm.ExtModel()
    ForcingModel_object = hcdfm.ForcingModel()
    boundary_object = hcdfm.Boundary(quantity='waterlevelbnd', #the FM quantity for tide is also waterlevelbnd
                                     locationfile=file_pli,
                                     forcingfile=ForcingModel_object)
    ext_new.boundary.append(boundary_object)
    ext_new.save(filepath=ext_file_new)
    
    # old ext
    ext_file_old = os.path.join(tmp_path, 'test_old.ext')
    ext_old = hcdfm.ExtOldModel()
    forcing_old = hcdfm.ExtOldForcing(quantity='airpressure_windx_windy_charnock',
                                      filename=file_pli,
                                      filetype=hcdfm.ExtOldFileType.NetCDFGridData, #11
                                      method=hcdfm.ExtOldMethod.InterpolateTimeAndSpaceSaveWeights, #3
                                      operand=hcdfm.Operand.override) #O
    ext_old.forcing.append(forcing_old)
    ext_old.save(filepath=ext_file_old)
    
    # initialize mdu file, update settings and save
    mdu_file = os.path.join(tmp_path, 'test_model.mdu')
    mdu = hcdfm.FMModel()
    mdu.geometry.fixedweirfile = file_pli
    mdu.external_forcing.extforcefile = ext_old
    mdu.external_forcing.extforcefilenew = ext_new
    mdu.save(mdu_file)
    
    # make relative and test
    dfmt.make_paths_relative(mdu_file)
    
    # assertions: including equal sign is essential to check for relative paths
    with open(mdu_file, 'r') as file:
        filedata = file.read()
    assert "= test_model.pli" in filedata
    assert "= test_old.ext" in filedata
    assert "= test_new.ext" in filedata
    
    with open(ext_file_old, 'r') as file:
        filedata = file.read()
    assert "FILENAME=test_model.pli" in filedata
    
    with open(ext_file_new, 'r') as file:
        filedata = file.read()
    assert "locationFile = test_model.pli" in filedata


@pytest.mark.unittest
def test_preprocess_merge_meteofiles_era5(tmp_path, ds_era5_empty):
    file_nc = os.path.join(tmp_path,"era5_msl_empty.nc")
    ds_era5_empty.to_netcdf(file_nc)
    
    ext_old = hcdfm.ExtOldModel()
    date_min = ds_era5_empty.time.to_pandas().iloc[0]
    date_max = ds_era5_empty.time.to_pandas().iloc[-1]
    varlist_list = ['msl']
    ext_old = dfmt.preprocess_merge_meteofiles_era5(
        ext_old=ext_old,
        varkey_list=varlist_list,
        dir_data=tmp_path,
        dir_output=tmp_path,
        time_slice=slice(date_min, date_max),
        )


@pytest.mark.unittest
def test_preprocess_merge_meteofiles_era5_unsupported_varlist(tmp_path, ds_era5_empty):
    file_nc = os.path.join(tmp_path,"era5_msl_empty.nc")
    ds_era5_empty.to_netcdf(file_nc)
    
    ext_old = hcdfm.ExtOldModel()
    date_min = ds_era5_empty.time.to_pandas().iloc[0]
    date_max = ds_era5_empty.time.to_pandas().iloc[-1]
    varlist_list = [['msl']]
    with pytest.raises(TypeError) as e:
        ext_old = dfmt.preprocess_merge_meteofiles_era5(ext_old=ext_old,
                                                        varkey_list=varlist_list,
                                                        dir_data=tmp_path,
                                                        dir_output=tmp_path,
                                                        time_slice=slice(date_min, date_max))
    error_msg = (
        "varkey_list should not contain lists, support was dropped in favour of "
        "supporting separate quantities. Provide a list of strings instead")
    assert error_msg in str(e.value)


@pytest.mark.unittest
def test_preprocess_merge_meteofiles_era5_unsupported_varkey(tmp_path, ds_era5_empty):
    file_nc = os.path.join(tmp_path,"era5_msl_empty.nc")
    ds_era5_empty.to_netcdf(file_nc)
    
    ext_old = hcdfm.ExtOldModel()
    date_min = ds_era5_empty.time.to_pandas().iloc[0]
    date_max = ds_era5_empty.time.to_pandas().iloc[-1]
    varlist_list = ['msla']
    with pytest.raises(NotImplementedError) as e:
        ext_old = dfmt.preprocess_merge_meteofiles_era5(ext_old=ext_old,
                                                        varkey_list=varlist_list,
                                                        dir_data=tmp_path,
                                                        dir_output=tmp_path,
                                                        time_slice=slice(date_min, date_max))
    assert "The varkey 'msla' is not supported yet by" in str(e.value)


@pytest.mark.unittest
def test_preprocess_merge_meteofiles_era5_missing_files(tmp_path, ds_era5_empty):
    file_nc = os.path.join(tmp_path,"era5_msl_empty.nc")
    ds_era5_empty.to_netcdf(file_nc)
    
    ext_old = hcdfm.ExtOldModel()
    date_min = ds_era5_empty.time.to_pandas().iloc[0]
    date_max = ds_era5_empty.time.to_pandas().iloc[-1]
    varlist_list = ['rhoao']
    with pytest.raises(FileNotFoundError) as e:
        ext_old = dfmt.preprocess_merge_meteofiles_era5(ext_old=ext_old,
                                                        varkey_list=varlist_list,
                                                        dir_data=tmp_path,
                                                        dir_output=tmp_path,
                                                        time_slice=slice(date_min, date_max))
    assert "No files found for pattern" in str(e.value)



import numpy as np
import xarray as xr


def write_bathy_toasc(filename_asc,lon_sel_ext,lat_sel_ext,elev_sel_ext,asc_fmt='%9.2f',nodata_val=32767):
    """
    empty docstring
    """

    print('writing to asc file')
    if elev_sel_ext.shape != (lat_sel_ext.shape[0], lon_sel_ext.shape[0]):
        raise ValueError('resulting shape of elev_sel_ext %s is not consistent with lat_sel_ext/lon_sel_ext vars %s'%(elev_sel_ext.shape,(lat_sel_ext.shape[0], lon_sel_ext.shape[0])))
    if isinstance(elev_sel_ext,np.ma.core.MaskedArray): #masked has to be filled in order for the nans to be visible
        elev_sel_ext = elev_sel_ext.filled(np.nan)
    if np.isnan(elev_sel_ext).sum()>0:
        elev_sel_ext = elev_sel_ext.copy()
        elev_sel_ext[np.isnan(elev_sel_ext)] = nodata_val
        print('replaced nan values with %d'%(nodata_val))
    resinv_lonlat = np.round(1/np.diff(lon_sel_ext[:2]),2)[0]
    resinv_lat = np.round(1/np.diff(lat_sel_ext[:2]),2)[0]
    if resinv_lonlat!=resinv_lat:
        raise ValueError('inconsistent resolution over lat/lon')
    
    with open(filename_asc,'w') as file_asc:
        file_asc.write('ncols         %d\n'%(elev_sel_ext.shape[1]))
        file_asc.write('nrows         %d\n'%(elev_sel_ext.shape[0]))
        file_asc.write('xllcenter     %13.8f\n'%(lon_sel_ext[0]))
        file_asc.write('yllcenter     %13.8f\n'%(lat_sel_ext[0]))
        file_asc.write('cellsize      %16.11f\n'%(1/resinv_lonlat))
        #file_asc.write('NODATA_value  %d\n'%(nodata_val))
        file_asc.write('NODATA_value  '+asc_fmt%(nodata_val)+'\n')
    with open(filename_asc,'a') as file_asc:
        np.savetxt(file_asc,np.flip(elev_sel_ext,axis=0),fmt=asc_fmt)
    print('...finished')


def read_asc(file_asc:str) -> xr.Dataset:
    """
    Reading asc file into a xarray.Dataset

    Parameters
    ----------
    file_asc : str
        asc file with header ncols, nrows, xllcenter, yllcenter, cellsize, NODATA_value.

    Returns
    -------
    ds_asc : xr.Dataset
        xarray.Dataset with the data from the asc file as an array and 
        the lat/lon coordinates as separate coordinate variables.

    """
    # read header
    header_dict = {}
    with open(file_asc) as f:
        for linenum, line in enumerate(f, 1):
            linesplit = line.split()
            linestart = linesplit[0]
            linestop = linesplit[-1]
            
            try:
                # try to convert it to float, this will fail for headerlines
                # it will succeed for numeric lines, but then the loop breaks
                float(linestart)
                break
            except ValueError:
                # convert header values to int if possible or float if not
                try:
                    header_value = int(linestop)
                except ValueError:
                    header_value = float(linestop)
                header_dict[linestart] = header_value
                skiprows = linenum
    
    # read data
    asc_np = np.loadtxt(file_asc, skiprows=skiprows, dtype=float)
    
    # derive x/y values and assert shape
    num_x = header_dict['ncols']
    num_y = header_dict['nrows']
    start_x = header_dict['xllcenter']
    start_y = header_dict['yllcenter']
    step = header_dict['cellsize']
    nodata = header_dict['NODATA_value']
    
    assert asc_np.shape == (num_y, num_x)
    
    x_vals = np.arange(0, num_x) * step + start_x
    y_vals = np.arange(0, num_y) * step + start_y
    
    # flip over latitude and replace nodata with nan
    asc_np = np.flipud(asc_np)
    asc_np[asc_np==nodata] = np.nan
    
    ds_asc = xr.Dataset()
    ds_asc['lon'] = xr.DataArray(x_vals, dims=('lon'))
    ds_asc['lat'] = xr.DataArray(y_vals, dims=('lat'))
    ds_asc['data'] = xr.DataArray(asc_np, dims=('lat','lon'))
    ds_asc = ds_asc.assign_attrs(header_dict)
    return ds_asc




# -*- coding: utf-8 -*-
"""
Created on Mon Jun 19 22:03:43 2023

@author: veenstra
"""

import os
import pytest
import xarray as xr
import xugrid as xu
import numpy as np
import matplotlib.pyplot as plt
import dfm_tools as dfmt
import contextily as ctx
from requests import HTTPError


@pytest.mark.unittest
def test_import_shapely():
    """
    tests whether shapely can be imported successfully, this is a problem in some environments
    in that case 'import shapely' works, but import 'shapely.geometry' fails
    """
    import shapely
    import shapely.geometry


@pytest.mark.unittest
def test_import_numba():
    """
    for some reason, numba often fails importing if versions of numpy/scipy are a bit too new
    """
    import numba as nb

    
@pytest.mark.unittest
def test_modplot_velovect():
    """
    this test will fail with matplotlib<3.6.0
    """
    x = np.linspace(-4,4,120)
    y = np.linspace(-3,3,100)
    X,Y = np.meshgrid(x,y)
    U = -1 - X**2 + Y
    V = 1 + X - Y**2
    speed = np.sqrt(U*U + V*V)
    grains = 15
    
    # dfmt.velovect requires matplotlib>=3.4.0
    fig,ax = plt.subplots()
    dfmt.velovect(ax, X, Y, U, V, color=speed, cmap='winter', arrowstyle='fancy', 
                  linewidth=speed/5, integration_direction='forward',
                  density=5, grains=grains)


@pytest.mark.unittest
def test_matplotlib_streamplot_broken_streamlines():
    """
    this test will fail with matplotlib<3.6.0
    """
    x = np.linspace(-4,4,120)
    y = np.linspace(-3,3,100)
    X,Y = np.meshgrid(x,y)
    U = -1 - X**2 + Y
    V = 1 + X - Y**2
    speed = np.sqrt(U*U + V*V)
    
    # broken_streamlines requires matplotlib>=3.6.0
    fig,ax = plt.subplots()
    ax.streamplot(X, Y, U, V, color=speed, cmap='winter', arrowstyle='fancy', linewidth=speed/5, broken_streamlines=False)


@pytest.mark.unittest
def test_xugrid_opendataset_ugridplot():
    """
    this one used to fail with xarray>=2023.3.0: https://github.com/Deltares/xugrid/issues/78
    This will probably not happen again, but it is convenient to test anyway.
    """
    file_nc = dfmt.data.fm_curvedbend_map(return_filepath=True)
    uds = xu.open_dataset(file_nc,chunks={'time':1})
    uds['mesh2d_flowelem_bl'].ugrid.plot()


@pytest.mark.unittest
def test_xugrid_opendataset_ugridplot_contourf():
    """
    This testcase gave a DeprecationWarning with scipy<1.10.0: https://github.com/Deltares/dfm_tools/issues/557
    Contourf/contour resulted in several other issues,
    so it is useful to test if the function can be called
    this test will fail with matplotlib 3.4.0 and 3.5.0
    """
    file_nc = dfmt.data.fm_curvedbend_map(return_filepath=True)
    uds = xu.open_dataset(file_nc)
    uds['mesh2d_flowelem_bl'].ugrid.plot.contourf()


@pytest.mark.unittest
def test_xugrid_opendataset_ugridplot_contour_with_colorbar():
    """
    Plotting a contour plot on uniform data (bedlevel -10 meter everywhere),
    resulted in several errors when adding a colorbar with matplotlib<=3.6.0
    """
    file_nc = dfmt.data.fm_curvedbend_map(return_filepath=True)
    uds = xu.open_dataset(file_nc)
    uds['mesh2d_flowelem_bl'].ugrid.plot.contour(add_colorbar=True)


@pytest.mark.unittest
def test_xarray_pandas_resample():
    """
    this one fails with xarray<2023.3.0 (required for test_opendataset_ugridplot()) and pandas>1.5.3: https://github.com/Deltares/xugrid/issues/78#issuecomment-1597723955
    pandas<1.5.3 works with xarray (any version)
    pandas>=2.0.0 fails with xarray<2023.3.0
    pandas>=2.0.0 cannot be installed with xarray==2023.3.0 (latter requires pandas<2 and >=1.4)
    pandas>=2.0.0 works with xarray>=2023.4.0
    therefore, xarray>=2023.4.0 works pandas (any version)
    """
    ds = xr.tutorial.load_dataset("air_temperature")
    ds.resample(time='D')

    
@pytest.mark.unittest
def test_xarray_decode_default_fillvals(tmp_path):
    """
    This test will fail as soon as xarray handles default fillvalues: https://github.com/Deltares/dfm_tools/issues/490
    After that, the minimum xarray requirement can be updated
    However, py38 support must then be dropped: https://github.com/Deltares/dfm_tools/issues/267
    In that case, this testcase and `dfmt.decode_default_fillvals()` can be removed 
    """
    
    import dfm_tools as dfmt
    import xarray as xr
    from netCDF4 import default_fillvals
    
    file_nc = dfmt.data.fm_grevelingen_map(return_filepath=True)
    file_nc = file_nc.replace('_0*','_0002')
    
    ds = xr.open_dataset(file_nc,decode_cf=False)
    
    #convert fillvalue in fnc to default fillvalue
    varn_fnc = 'mesh2d_face_nodes'
    fnc_dtype = ds[varn_fnc].dtype.str[1:]
    fill_value_default = default_fillvals[fnc_dtype]
    fill_value = ds[varn_fnc].attrs.pop('_FillValue')
    ds[varn_fnc] = ds[varn_fnc].where(ds[varn_fnc]!=fill_value,fill_value_default)
    
    #write file
    file_out = os.path.join(tmp_path, 'temp_fnc_default_fillvals_map.nc')
    ds.to_netcdf(file_out)
    
    #open dataset with decode_fillvals
    with pytest.raises(ValueError) as e:
        # until xarray handles default fillvalues: https://github.com/Deltares/dfm_tools/issues/490
        # if xarray handles default fillvalues, `dfmt.decode_default_fillvals()` can be removed if minimum xarray version is set as requirement
        dfmt.open_partitioned_dataset(file_out, decode_fillvals=False)
    assert "connectivity contains negative values" in str(e.value)

    #this should be successful
    uds2 = dfmt.open_partitioned_dataset(file_out, decode_fillvals=True)
    fnc_new = uds2.grid.face_node_connectivity
    
    assert -1 in fnc_new


@pytest.mark.unittest
def test_ctx_add_basemap():
    """
    tests wheter this often used function still works, it fails in contextily 1.3.0
    mainly relevant to update the minimal contextily version in the pyproject.toml file sometimes
    https://github.com/Deltares/dfm_tools/issues/857
    """
    _, ax = plt.subplots()
    ax.set_xlim(1.8,3.2)
    ax.set_ylim(50.8, 52.2)
    try:
        ctx.add_basemap(ax=ax, crs="EPSG:4326", attribution=False)
    except HTTPError as e:
        # servers are flaky so HTTPError is raised quite often
        # this should not result in a failing dfm_tools test since it is just temporary
        print(e)
    except AttributeError as e:
        # can be removed after fixing https://github.com/geopandas/contextily/issues/252
        assert str(e) == "'NoneType' object has no attribute 'shape'"
        print(e)
    except TypeError as e:
        # can be removed after fixing https://github.com/geopandas/contextily/issues/252
        assert "not 'NoneType'" in str(e)
        print(e)



def get_ncmodeldata(file_nc=None, varname=None, timestep=None, layer=None, depth=None, station=None, multipart=None, get_linkedgridinfo=False, silent=False):
    raise DeprecationWarning('dfmt.get_ncmodeldata() is deprecated, use `uds = dfmt.open_partitioned_dataset(file_nc)` instead: https://github.com/Deltares/dfm_tools/blob/main/notebooks/postprocessing_example.ipynb')


def get_netdata(file_nc=None, multipart=None):
    raise DeprecationWarning('dfmt.get_netdata() is deprecated, use `uds = dfmt.open_partitioned_dataset(file_nc)` instead: https://github.com/Deltares/dfm_tools/blob/main/notebooks/postprocessing_example.ipynb')


def plot_netmapdata(verts=None, values=None, ax=None, **kwargs):
    raise DeprecationWarning('dfmt.plot_netmapdata() is deprecated, use `uds = dfmt.open_partitioned_dataset(file_nc); uds.mesh2d_s1.isel(time=0).ugrid.plot()` instead: https://github.com/Deltares/dfm_tools/blob/main/notebooks/postprocessing_example.ipynb')


def plot_background(ax=None, projection=None, google_style='satellite', resolution=1, features=None, nticks=6, latlon_format=False, gridlines=False, **kwargs):
    raise DeprecationWarning('dfmt.plot_background() is deprecated, use contextily instead')


def get_ugrid_verts(data_xr_map=None):
    raise DeprecationWarning('dfmt.get_ugrid_verts() is deprecated, use uds.grid.face_node_coordinates instead (https://github.com/Deltares/xugrid/issues/48)')


def scatter_to_regulargrid(xcoords=None, ycoords=None, values=None, ncellx=None, ncelly=None, reg_x_vec=None, reg_y_vec=None, method='nearest', maskland_dist=None):
    raise DeprecationWarning('dfm_tools.regulargrid.scatter_to_regulargrid() is deprecated, use ds = dfmt.rasterize_ugrid(uds) instead')


def get_varnamefromattrs(data_xr=None, varname=None):
    raise DeprecationWarning('dfmt.get_varnamefromattrs() will be deprecated in a future version of dfm_tools, ds=dfmt.rename_waqvars(ds) is a more convenient alternative')


class Polygon:
    def __init__(self, data=None, name=None, comments=None):
        raise DeprecationWarning('the function dfm_tools.polygon.Polygon() is deprecated, please use the new hydrolib alternative.')
        
    def fromfile(self, file_pol=None, pd_output=False, tekmap_output=False):
        raise DeprecationWarning('the function dfm_tools.polygon.Polygon.fromfile() is deprecated, please use the new hydrolib alternative. Example script: https://github.com/Deltares/dfm_tools/blob/main/tests/examples/preprocess_hydrolib_readwritepol.py')


def write_bcfile(filename=None, datablocks=None, metadatas=None, refdate=None, tzone=0, float_format='%6.2f'):
    raise DeprecationWarning('the function dfm_tools.io.bc.write_bcfile() is deprecated, please use the new hydrolib alternative. Example script: dfm_tools/tests/examples/CMEMS_interpolate_example.py')


def read_bcfile(filename=None, converttime=False):
    raise DeprecationWarning('the function dfm_tools.io.bc.read_bcfile() is deprecated, please use the new hydrolib alternative. Example script: dfm_tools/tests/examples/hydrolib_readbc.py')


def write_timfile(filename=None, datablock=None, header=None, converttime=False, refdate=None, float_format='%6.2f'):
    raise DeprecationWarning('the function dfm_tools.write_timfile() is deprecated, please use the new hydrolib alternative: https://github.com/Deltares/dfm_tools/blob/301-convert-timmodel-to-pandasdataframe/tests/examples/preprocess_hydrolib_readtim.py.')


def read_timfile(filename=None, converttime=False, refdate=None):
    raise DeprecationWarning('the function dfm_tools.read_timfile() is deprecated, please use the new hydrolib alternative: https://github.com/Deltares/dfm_tools/blob/301-convert-timmodel-to-pandasdataframe/tests/examples/preprocess_hydrolib_readtim.py.')


def generate_bndpli(**kwargs):
    raise DeprecationWarning('the function dfmt.generate_bndpli() is deprecated, please use dfmt.generate_bndpli_cutland() instead.')


def preprocess_hirlam(ds=None):
    raise DeprecationWarning('the function dfmt.preprocess_hirlam() is deprecated, xarray now supports datasets with multidimensional coordinates.')


def interp_regularnc_to_plipoints(ds=None):
    raise DeprecationWarning('the function dfmt.interp_regularnc_to_plipoints() is deprecated, '
                             'use dfmt.interp_regularnc_to_plipointsDataset() instead with gdf_points '
                             'as in https://github.com/Deltares/dfm_tools/issues/938')

def open_dataset_extra(**kwargs):
    raise DeprecationWarning('the function dfmt.open_dataset_extra() is deprecated, dfmt.open_prepare_dataset() is similar but does not support multiple quantities at once')


def preprocess_ini_cmems_to_nc(**kwargs):
    raise DeprecationWarning("`dfmt.preprocess_ini_cmems_to_nc()` was "
                             "deprecated, use `cmems_nc_to_ini()` instead")



import xugrid as xu
import meshkernel
import xarray as xr
import datetime as dt
from dfm_tools import __version__
import getpass
import numpy as np
from dfm_tools.coastlines import get_coastlines_gdb
from netCDF4 import default_fillvals
import geopandas as gpd
from shapely import MultiPolygon, LineString, MultiLineString
from shapely.ops import linemerge
from itertools import groupby
from shapely import Polygon

__all__ = [
    "meshkernel_delete_withcoastlines",
    "meshkernel_delete_withshp",
    "meshkernel_delete_withgdf",
    "meshkernel_get_illegalcells",
    "meshkernel_to_UgridDataset",
    "meshkernel_get_bbox",
    "make_basegrid",
    "refine_basegrid",
    "generate_bndpli_cutland",
    "interpolate_bndpli",
    ]

def meshkernel_delete_withcoastlines(mk:meshkernel.MeshKernel, res:str='f', min_area:float = 0, crs:(int,str) = None):
    """
    Wrapper around meshkernel_delete_withgdf, which automatically gets the bbox from the meshkernel object and retrieves the coastlines_gdf.

    Parameters
    ----------
    mk : meshkernel.MeshKernel
        DESCRIPTION.
    res : str, optional
        DESCRIPTION. The default is 'f'.
    min_area : float, optional
        DESCRIPTION. The default is 0.
    crs : (int,str), optional
        DESCRIPTION. The default is None.

    Returns
    -------
    None.

    """
    
    mesh_bnds = mk.mesh2d_get_mesh_boundaries_as_polygons()
    
    bbox = (mesh_bnds.x_coordinates.min(), mesh_bnds.y_coordinates.min(), mesh_bnds.x_coordinates.max(), mesh_bnds.y_coordinates.max())
    
    coastlines_gdf = get_coastlines_gdb(bbox=bbox, res=res, min_area=min_area, crs=crs)
    
    meshkernel_delete_withgdf(mk, coastlines_gdf)


def meshkernel_delete_withshp(mk:meshkernel.MeshKernel, coastlines_shp:str, crs:(int,str) = None):
    """
    Delete parts of mesh that are inside the shapefile polygon.

    Parameters
    ----------
    mk : meshkernel.MeshKernel
        DESCRIPTION.
    coastlines_shp : str
        Path to the shp file.
    crs : (int,str), optional
        DESCRIPTION. The default is None.

    Returns
    -------
    None.
    
    """
        
    mesh_bnds = mk.mesh2d_get_mesh_boundaries_as_polygons()
    bbox = (mesh_bnds.x_coordinates.min(), mesh_bnds.y_coordinates.min(), mesh_bnds.x_coordinates.max(), mesh_bnds.y_coordinates.max())
    coastlines_gdb = gpd.read_file(coastlines_shp, bbox=bbox)
    
    if crs:
        coastlines_gdb = coastlines_gdb.to_crs(crs)
    
    meshkernel_delete_withgdf(mk, coastlines_gdb)


def meshkernel_delete_withgdf(mk:meshkernel.MeshKernel, coastlines_gdf:gpd.GeoDataFrame):
    """
    Delete parts of mesh that are inside the polygons/Linestrings in a GeoDataFrame.

    Parameters
    ----------
    mk : meshkernel.MeshKernel
        DESCRIPTION.
    coastlines_gdf : gpd.GeoDataFrame
        DESCRIPTION.

    Returns
    -------
    None.

    """
    
    for coastline_geom in coastlines_gdf['geometry']:
        xx, yy = coastline_geom.exterior.coords.xy
        xx = np.array(xx)
        yy = np.array(yy)
        
        delete_pol_geom = meshkernel.GeometryList(x_coordinates=xx, y_coordinates=yy) #TODO: .copy()/to_numpy() makes the array contiguous in memory, which is necessary for meshkernel.mesh2d_delete()
        mk.mesh2d_delete(geometry_list=delete_pol_geom, 
                         delete_option=meshkernel.DeleteMeshOption.INSIDE_NOT_INTERSECTED,
                         invert_deletion=False)


def meshkernel_get_illegalcells(mk):
    # get illegalcells from meshkernel instance
    illegalcells_geom = mk.mesh2d_get_face_polygons(num_edges=6)
    # convert xy coords to numpy array
    illegalcells_np = np.c_[illegalcells_geom.x_coordinates, illegalcells_geom.y_coordinates]
    # split illegalcells array based on the geomtry_separator
    xy_lists = [list(g) for k, g in groupby(illegalcells_np, lambda x: (x != illegalcells_geom.geometry_separator).all()) if k]
    # convert to geodataframe of Polygons
    list_polygons = [Polygon(xylist) for xylist in xy_lists]
    illegalcells_gdf = gpd.GeoDataFrame(geometry=list_polygons)
    return illegalcells_gdf


def geographic_to_meshkernel_projection(is_geographic:bool) -> meshkernel.ProjectionType:
    """
    converts is_geographic boolean to meshkernel.ProjectionType (SPHERICAL OR CARTESIAN)

    Parameters
    ----------
    is_geographic : bool
        DESCRIPTION.

    Returns
    -------
    projection : TYPE
        DESCRIPTION.

    """
    if is_geographic:
        projection = meshkernel.ProjectionType.SPHERICAL
    else:
        projection = meshkernel.ProjectionType.CARTESIAN
    return projection


def meshkernel_is_geographic(mk):
    if mk.get_projection()==meshkernel.ProjectionType.CARTESIAN:
        is_geographic = False
    else:
        is_geographic = True
    return is_geographic


def meshkernel_to_UgridDataset(mk:meshkernel.MeshKernel, crs:(int,str) = None) -> xu.UgridDataset:
    """
    Convert a meshkernel object to a UgridDataset, including a variable with the crs (used by dflowfm to distinguish spherical/cartesian networks).
    The UgridDataset enables bathymetry interpolation and writing to netfile.

    Parameters
    ----------
    mk : meshkernel.MeshKernel
        DESCRIPTION.
    crs : (int,str), optional
        DESCRIPTION. The default is None.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    
    crs_is_geographic = crs_to_isgeographic(crs)
    
    mesh2d_grid = mk.mesh2d_get()
    
    #check if both crs and grid are geograpic or not
    #TODO: do this in xugrid: https://github.com/Deltares/xugrid/issues/188
    grid_is_geographic = meshkernel_is_geographic(mk)
    if crs_is_geographic != grid_is_geographic:
        raise ValueError(f"crs has is_geographic={crs_is_geographic} and grid has is_geographic={grid_is_geographic}. This is not allowed.")
    
    # TODO: below is not correctly handled by xugrid yet, projected=False does not give is_geographic=True
    # related issue is https://github.com/Deltares/xugrid/issues/187
    xu_grid = xu.Ugrid2d.from_meshkernel(mesh2d_grid, projected= not crs_is_geographic, crs=crs)
    
    # convert 0-based to 1-based indices for connectivity variables like face_node_connectivity
    # this is required by delft3dfm
    xu_grid.start_index = 1
    xu_grid_ds = xu_grid.to_dataset()
    
    # convert to uds and add attrs and crs
    xu_grid_uds = xu.UgridDataset(xu_grid_ds)
    
    xu_grid_uds = xu_grid_uds.assign_attrs({#'Conventions': 'CF-1.8 UGRID-1.0 Deltares-0.10', #TODO: conventions come from xugrid, so this line is probably not necessary
                                          'institution': 'Deltares',
                                          'references': 'https://www.deltares.nl',
                                          'source': f'Created with meshkernel {meshkernel.__version__}, xugrid {xu.__version__} and dfm_tools {__version__}',
                                          'history': 'Created on %s, %s'%(dt.datetime.now().strftime('%Y-%m-%dT%H:%M:%S%z'),getpass.getuser()), #TODO: add timezone
                                          })
    
    # add crs including attrs
    if crs is not None:
        xu_grid_uds.ugrid.set_crs(crs)
        uds_add_crs_attrs(xu_grid_uds)
    return xu_grid_uds


def uds_get_crs(uds):
    uds_crs_dict = uds.ugrid.crs
    if uds_crs_dict is None:
        return None
    else:
        keys = list(uds_crs_dict.keys())
        crs = uds_crs_dict[keys[0]]
        return crs


def uds_add_crs_attrs(uds:(xu.UgridDataset,xr.Dataset)):
    """
    

    Parameters
    ----------
    uds : (xu.UgridDataset,xr.Dataset)
        DESCRIPTION.
    crs : (str,int)
        epsg, e.g. 'EPSG:4326' or 4326.

    Raises
    ------
    ValueError
        DESCRIPTION.

    Returns
    -------
    None.

    """
    # TODO: upon crs conversion and to_netcdf(), the netcdf file could get two crs vars, catch this
    
    grid_is_geographic = uds.grid.is_geographic
    
    crs = uds_get_crs(uds)
    
    #get crs information (name/num)
    if crs is None:
        crs_num = 0
        crs_str = 'EPSG:0'
        crs_name = ''
        crs_is_geographic = False
    else:
        # get crs info, should actually be `import pyproj; pyproj.CRS.from_user_input(crs)`
        crs_info = gpd.GeoSeries(crs=crs).crs #also contains area-of-use (name/bounds), datum (ellipsoid/prime-meridian)
        crs_num = crs_info.to_epsg()
        crs_str = crs_info.to_string()
        crs_name = crs_info.name
        crs_is_geographic = crs_info.is_geographic
        # TODO: standard names for lat/lon in crs_info.cs_to_cf()
    
    #check if combination of is_geographic and crs makes sense
    if grid_is_geographic != crs_is_geographic:
        raise ValueError(f"`grid_is_geographic` mismatch between provided grid (is_geographic={grid_is_geographic}) and provided crs ({crs}, is_geographic={crs_is_geographic})")
    
    # TODO: consider always using the same crs_varn, align with xugrid
    # QGIS also does not recognize epsg anymore when renaming variable to `crs` 
    # or something else (`wgs84` and `projected_coordinate_system` both do work)
    if grid_is_geographic:
        crs_varn = 'wgs84'
    else:
        crs_varn = 'projected_coordinate_system'
    
    attribute_dict = {
        'name': crs_name, # not required, but convenient for the user
        'epsg': np.array(crs_num, dtype=int), # epsg or EPSG_code should be present for the interacter to load the grid and by QGIS to recognize the epsg.
        'EPSG_code': crs_str, # epsg or EPSG_code should be present for the interacter to load the grid and by QGIS to recognize the epsg.
        }
    if grid_is_geographic:
        # without grid_mapping_name="latitude_longitude", interacter sees the grid as cartesian
        # grid_mapping_name is not available for projected crs's
        attribute_dict['grid_mapping_name'] = crs_info.to_cf()['grid_mapping_name']
    
    uds[crs_varn] = xr.DataArray(np.array(default_fillvals['i4'],dtype=int),dims=(),attrs=attribute_dict)


def crs_to_isgeographic(crs=None):
    if crs is None:
        is_geographic = False
    else:
        import pyproj
        crs = pyproj.CRS.from_user_input(crs)
        is_geographic = crs.is_geographic
    return is_geographic


def make_basegrid(lon_min,lon_max,lat_min,lat_max,dx,dy,angle=0,
                  crs=None, is_geographic=None):
    """
    empty docstring
    """
    
    if is_geographic is not None:
        raise ValueError("'is_geographic' was deprecated as argument for dfmt.make_basegrid(), it is now derived from 'crs' instead")
    
    
    is_geographic = crs_to_isgeographic(crs)
    projection = geographic_to_meshkernel_projection(is_geographic)
    
    # create base grid
    make_grid_parameters = meshkernel.MakeGridParameters(angle=angle,
                                                         origin_x=lon_min,
                                                         origin_y=lat_min,
                                                         upper_right_x=lon_max,
                                                         upper_right_y=lat_max,
                                                         block_size_x=dx,
                                                         block_size_y=dy)
    
    mk = meshkernel.MeshKernel(projection=projection)
    mk.curvilinear_compute_rectangular_grid_on_extension(make_grid_parameters)
    mk.curvilinear_convert_to_mesh2d() #convert to ugrid/mesh2d
    
    return mk


def refine_basegrid(
        mk:meshkernel.MeshKernel,
        data_bathy_sel:xr.DataArray,
        **kwargs,
        ):
    """
    Refine the grid based on gridded bathymetry.

    Parameters
    ----------
    mk : meshkernel.MeshKernel
        The meshkernel instance containing the base grid. Updated in place.
    data_bathy_sel : xr.DataArray
        The bathymetry data used for the refinement. Is converted to 
        meshkernel.GriddedSamples().
    **kwargs : TYPE
        Arguments passed on to meshkernel.MeshRefinementParameters(). Some
        options from the meshkernelpy docs:
        * min_edge_size (float): Minimum edge size. Meshkernelpy default is
        `0.5`. Overwritten with 500 here.
        * refinement_type (RefinementType): Refinement criterion type. 
        Meshkernelpy default is `RefinementType.REFINEMENT_LEVELS`. Overwritten
        with `RefinementType.WAVE_COURANT` here.
        * smoothing_iterations (int, optional): The number of smoothing
        iterations. Meshkernelpy default is `5`. Overwritten with 2 here.
        * connect_hanging_nodes (bool): Whether to connect hanging nodes at the
        end of the iteration. Meshkernelpy default is `True`. Set to False to
        do multiple refinement steps (e.g. for multiple regions)
        * max_courant_time (double, optional): Maximum courant time in seconds.
        Meshkernelpy default is `120`.
        
    """
    
    if 'min_edge_size' not in kwargs:
        # min_edge_size (float): Minimum edge size. Meshkernelpy default is 
        # `0.5`. Overwrite with more sensible value for coastal models.
        kwargs['min_edge_size'] = 500
    if 'refinement_type' not in kwargs:
        # Refinement criterion type. Meshkernelpy default is 
        # `RefinementType.REFINEMENT_LEVELS`. 
        kwargs['refinement_type'] = meshkernel.RefinementType.WAVE_COURANT
    if 'smoothing_iterations' not in kwargs:
        # The number of smoothing iterations. Meshkernelpy default is `5`.
        kwargs['smoothing_iterations'] = 2

    lon_np = data_bathy_sel.lon.to_numpy()
    lat_np = data_bathy_sel.lat.to_numpy()
    values_np = data_bathy_sel.to_numpy().flatten()
    gridded_samples = meshkernel.GriddedSamples(
        x_coordinates=lon_np,
        y_coordinates=lat_np,
        values=values_np)

    #refinement
    mesh_refinement_parameters = meshkernel.MeshRefinementParameters(**kwargs)
    
    mk.mesh2d_refine_based_on_gridded_samples(
        gridded_samples=gridded_samples,
        mesh_refinement_params=mesh_refinement_parameters,
        use_nodal_refinement=True,
        )


def meshkernel_get_outer_xycoords(mk:meshkernel.MeshKernel):
    mesh_bnds = mk.mesh2d_get_mesh_boundaries_as_polygons()
    xcoords = mesh_bnds.x_coordinates
    ycoords = mesh_bnds.y_coordinates
    if mesh_bnds.geometry_separator in xcoords:
        # TODO: would be nice to get only the outer xycoords instead
        # or to have easier selection of outer xycoords
        raise Exception('use meshkernel_get_outer_xycoords() on an uncut grid')
    return xcoords, ycoords

    
def meshkernel_get_bbox(mk:meshkernel.MeshKernel):
    """
    get the bbox of a meshkernel instance as (xmin, ymin, xmax, ymax)
    """
    xcoords, ycoords = meshkernel_get_outer_xycoords(mk)
    bbox = (xcoords.min(), ycoords.min(), xcoords.max(), ycoords.max())
    return bbox


def generate_bndpli_cutland(mk:meshkernel.MeshKernel, res:str='f', min_area:float = 0, crs:(int,str) = None, buffer:float = 0):
    """
    Generate a boundary polyline from the meshkernel object and cut away the landward part.
    Be sure to do this on the base/refined grid, not on the grid where the landward cells were already cut.
    
    Parameters
    ----------
    mk : meshkernel.MeshKernel
        DESCRIPTION.
    res : str, optional
        DESCRIPTION. The default is 'f'.
    min_area : float, optional
        DESCRIPTION. The default is 0.
    crs : (int,str), optional
        DESCRIPTION. The default is None.
    buffer : float, optional
        DESCRIPTION. The default is 0.

    Returns
    -------
    bnd_gdf : TYPE
        DESCRIPTION.

    """
    bbox = meshkernel_get_bbox(mk)
    coastlines_gdf = get_coastlines_gdb(bbox=bbox, res=res, min_area=min_area, crs=crs)
    
    xcoords,ycoords = meshkernel_get_outer_xycoords(mk)
    mesh_bnds_xy = np.c_[xcoords,ycoords]
    meshbnd_ls = LineString(mesh_bnds_xy)
    coastlines_mp = MultiPolygon(coastlines_gdf.geometry.tolist())
    coastlines_mp = coastlines_mp.buffer(buffer)
    bnd_ls = meshbnd_ls.difference(coastlines_mp)
    
    #attempt to merge MultiLineString to single LineString
    if isinstance(bnd_ls,MultiLineString):
        print('attemting to merge lines in MultiLineString to single LineString (if connected)')
        bnd_ls = linemerge(bnd_ls)
    
    #convert MultiLineString/LineString to GeoDataFrame
    if isinstance(bnd_ls,MultiLineString):
        bnd_gdf = gpd.GeoDataFrame(geometry=list(bnd_ls.geoms))
    elif isinstance(bnd_ls,LineString):
        bnd_gdf = gpd.GeoDataFrame(geometry=[bnd_ls])
    
    #set crs from coastlines
    bnd_gdf.crs = coastlines_gdf.crs
    return bnd_gdf


def interpolate_bndpli(bnd_gdf,res):
    """
    interpolate bnd_gdf to a new resolution
    """
    #TODO: keep corners of grid, maybe use mk.polygon_refine()
    
    bnd_gdf_interp = bnd_gdf.copy()
    for irow,row in bnd_gdf_interp.iterrows():
        bnd_ls = row.geometry
        interp_range = np.arange(0,bnd_ls.length,res)
        bnd_ls_interp_points = bnd_ls.interpolate(interp_range)
        if len(bnd_ls_interp_points)==1: #no change if interp results in only one point
            continue
        bnd_ls_interp = LineString(bnd_ls_interp_points)
        bnd_gdf_interp.loc[irow,'geometry'] = bnd_ls_interp
    return bnd_gdf_interp




# -*- coding: utf-8 -*-
"""
Created on Sun Jul  9 22:35:28 2023

@author: veenstra
"""

import pytest
import geopandas as gpd
import matplotlib.pyplot as plt
from shapely import Polygon, LineString
import dfm_tools as dfmt


@pytest.mark.unittest
def test_get_coastlines_gdb_global():
    """
    opens global coarse coastlines at lower 30 degrees, so includes L6/Antarctic
    """
    min_area = 1000 # km2
    bbox = (-180, -90, 180, -60) # xmin, ymin, xmax, ymax
    res = 'c'
    gdf = dfmt.get_coastlines_gdb(res=res, min_area=min_area, bbox=bbox)
    
    assert len(gdf) == 17


@pytest.mark.unittest
def test_get_coastlines_gdb_allres():
    """
    opens coastline gdb of Ireland and UK, only islands > 1000km2
    """
    min_area = 1000 # km2
    bbox = (-10, 55, 0, 60) # xmin, ymin, xmax, ymax
    for res,ngeoms in zip('fhilc',[5,5,5,5,4]):
        gdf = dfmt.get_coastlines_gdb(res=res, min_area=min_area, bbox=bbox)
        assert len(gdf) == ngeoms
    assert isinstance(gdf,gpd.GeoDataFrame)
    assert isinstance(gdf.geometry[0],Polygon)
    assert 'area' in gdf.columns
    assert 'geometry' in gdf.columns


@pytest.mark.unittest
def test_get_coastlines_gdb_lakes_islandsinlakes():
    """
    opens coastline gdb of lake Kivu (so lake and island in lake, L2 and L3)
    """
    
    bbox = (28.9, -2.4, 29.3, -1.6) # xmin, ymin, xmax, ymax
    gdf = dfmt.get_coastlines_gdb(res='h', min_area=0, bbox=bbox)
    
    assert len(gdf) == 9


@pytest.mark.unittest
def test_plot_coastlines():
    
    fig, ax = plt.subplots()
    ax.set_xlim(-10,0)
    ax.set_ylim(50,60)
    dfmt.plot_coastlines(res='l', min_area=500)


@pytest.mark.unittest
def test_get_borders_gdb_global():
    """
    opens global coarse borders
    """
    bbox = (-180, -90, 180, 90) # xmin, ymin, xmax, ymax
    res = 'c'
    gdf = dfmt.get_borders_gdb(res=res, bbox=bbox)
    
    assert len(gdf) == 451


@pytest.mark.unittest
def test_get_borders_gdb_allres():
    """
    opens borders gdb of Ireland and UK
    """
    bbox = (-10, 40, 5, 60) # xmin, ymin, xmax, ymax
    for res in 'fhilc':
        gdf = dfmt.get_borders_gdb(res=res, bbox=bbox)
        assert len(gdf) == 9
    assert isinstance(gdf,gpd.GeoDataFrame)
    assert isinstance(gdf.geometry[0],LineString)
    assert 'geometry' in gdf.columns


@pytest.mark.unittest
def test_plot_borders():
    
    fig, ax = plt.subplots()
    ax.set_xlim(-10,5)
    ax.set_ylim(40,60)
    dfmt.plot_borders(res='l')



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 11 10:48:06 2023

@author: veenstra
"""

import os
import pytest
import numpy as np
import dfm_tools as dfmt


@pytest.mark.unittest
def test_read_write_asc(tmp_path):
    lat_range = np.arange(-8,2,0.25)
    lon_range = np.arange(-10,1,0.25)
    data = np.cos(lon_range) * lat_range[np.newaxis].T
    file_asc = os.path.join(tmp_path, 'dummy.asc')
    dfmt.write_bathy_toasc(file_asc, lon_sel_ext=lon_range, lat_sel_ext=lat_range, elev_sel_ext=data, asc_fmt='%14.9f',nodata_val=-999)
    
    ds_asc = dfmt.read_asc(file_asc)
    
    assert (np.abs(ds_asc['lon'].to_numpy() - lon_range) < 1e-9).all()
    assert (np.abs(ds_asc['lat'].to_numpy() - lat_range) < 1e-9).all()
    assert (np.abs(ds_asc['data'].to_numpy() - data) < 1e-9).all()
